import os
from functools import partial
from typing import Union, TYPE_CHECKING

import numpy as np
import pandas as pd

from mewpy.io.dto import VariableRecord, DataTransferObject, FunctionTerm
from mewpy.germ.algebra import Expression, And, Symbol
from mewpy.germ.models import RegulatoryModel
from .engine import Engine
from .engines_utils import csv_warning


if TYPE_CHECKING:
    from mewpy.germ.models import RegulatoryModel, Model, MetabolicModel


class CoExpressionRegulatoryCSV(Engine):

    def __init__(self, io, config, model=None):
        """
        Engine for CSV and TXT CoExpression (co-activating and co-repressing) regulatory files
        """
        super().__init__(io, config, model)

    @property
    def model_type(self):
        return 'regulatory'

    @property
    def model(self):

        if self._model is None:
            identifier = self.get_identifier()

            return RegulatoryModel(identifier=identifier)

        return self._model

    def build_data_frame(self):
        sep = self.config.get('sep', ',')
        target_col = self.config.get('target_col', 0)
        activating_regs = self.config.get('co_activating_col', 1)
        repressing_regs = self.config.get('co_repressing_col', 2)
        header = self.config.get('header', None)
        filter_nan = self.config.get('filter_nan', False)

        names = {target_col: 'targets', activating_regs: 'activating', repressing_regs: 'repressing'}

        try:
            df = pd.read_csv(self.io, sep=sep, header=header)

        except BaseException as exc:
            self.clean()
            self.close()
            raise exc

        cols = []

        for j, col in enumerate(df.columns):
            if j in names:
                cols.append(names[j])
            else:
                del df[col]

        df.columns = cols
        df.index = df.loc[:, 'targets']

        if filter_nan:

            df = df.dropna(subset=['activating', 'repressing'])

        else:

            df = df.replace(np.nan, '', regex=True)

        self.dto.data_frame = df

    def get_identifier(self):

        if os.path.exists(self.io):
            _, identifier = os.path.split(self.io)
            return os.path.splitext(identifier)[0]

    def open(self, mode='r'):

        self._dto = DataTransferObject()

        if not os.path.exists(self.io):
            raise OSError(f'{self.io} is not a valid input. Provide the path or file handler')

        self.dto.id = self.get_identifier()

    def parse(self):

        if self.dto is None:
            raise OSError('File is not open')

        if self.dto.id is None:
            raise OSError('File is not open')

        # -----------------------------------------------------------------------------
        # CSV/TXT to pandas dataframe
        # -----------------------------------------------------------------------------

        self.build_data_frame()

        for target in self.dto.data_frame.index:
            # -----------------------------------------------------------------------------
            # Target
            # -----------------------------------------------------------------------------
            target_id = target.replace(' ', '')

            target_aliases = self.dto.data_frame.loc[target, self.dto.aliases_columns]

            target_record = VariableRecord(id=target_id,
                                           name=target_id,
                                           aliases=set(target_aliases))

            self.variables[target_id].add('target')

            self.dto.targets[target_id] = target_record

            # -----------------------------------------------------------------------------
            # Regulators and Function terms
            # -----------------------------------------------------------------------------
            activating = self.dto.data_frame.loc[target, 'activating'].split(' ')
            repressing = self.dto.data_frame.loc[target, 'repressing'].split(' ')
            regulators = [repressing, activating]

            regulator_records = {}
            function_terms = {}
            for i, co_regulators in enumerate(regulators):

                variables = []

                for regulator in co_regulators:
                    self.variables[regulator].add('regulator')

                    regulator_record = VariableRecord(id=regulator,
                                                      name=regulator,
                                                      aliases={regulator})

                    regulator_records[regulator] = regulator_record

                    self.dto.regulators[regulator] = regulator_record

                    variables.append(Symbol(value=regulator))

                symbolic = And(variables=variables)

                function_terms[str(i)] = FunctionTerm(id=str(i), symbolic=symbolic, coefficient=i)

            # -----------------------------------------------------------------------------
            # Interaction
            # -----------------------------------------------------------------------------

            interaction_id = f'{target_id}_interaction'

            interaction_record = VariableRecord(id=interaction_id,
                                                name=interaction_id,
                                                aliases={target_id},
                                                target=target_record,
                                                function_terms=function_terms,
                                                regulators=regulator_records)

            self.dto.interactions[interaction_id] = interaction_record

            self.variables[interaction_id].add('interaction')

    def read(self,
             model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
             variables=None):

        if not model:
            model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = self.model

        if not variables:
            variables = self.variables

        if self.dto.id:
            model._id = self.dto.id

        processed_regulators = set()

        for interaction_id, interaction_record in self.dto.interactions.items():

            target_record = interaction_record.target

            target, warning = target_record.to_variable(model=model,
                                                        types=variables.get(target_record.id, {'target'}),
                                                        name=target_record.name,
                                                        aliases=target_record.aliases)

            if warning:
                self.warnings.append(partial(csv_warning, warning))

            regulators_records = interaction_record.regulators

            regulators = {}

            for regulator_id, regulator_record in regulators_records.items():

                regulator, warning = regulator_record.to_variable(model=model,
                                                                  types=variables.get(regulator_id, {'regulator'}),
                                                                  name=regulator_record.name,
                                                                  aliases=regulator_record.aliases)

                if warning:
                    self.warnings.append(partial(csv_warning, warning))

                regulators[regulator_id] = regulator

                processed_regulators.add(regulator_id)

            regulatory_events = {}

            for func_term in interaction_record.function_terms.values():
                expression_regulators = {symbol.name: regulators[symbol.name]
                                         for symbol in func_term.symbolic.atoms(symbols_only=True)}

                regulatory_events[func_term.coefficient] = Expression(symbolic=func_term.symbolic,
                                                                      variables=expression_regulators)

            interaction, warning = interaction_record.to_variable(model=model,
                                                                  types=variables.get(interaction_id, {'interaction'}),
                                                                  name=interaction_record.name,
                                                                  aliases=interaction_record.aliases,
                                                                  target=target,
                                                                  regulatory_events=regulatory_events)

            if warning:
                self.warnings.append(partial(csv_warning, warning))

            model.add(interaction, comprehensive=True)

        if len(processed_regulators) != len(self.dto.regulators):

            for regulator_id, regulator_record in self.dto.regulators.items():

                if regulator_id not in processed_regulators:

                    regulator, warning = regulator_record.to_variable(model=model,
                                                                      types=variables.get(regulator_id, {'regulator'}),
                                                                      name=regulator_record.name,
                                                                      aliases=regulator_record.aliases)

                    if warning:
                        self.warnings.append(partial(csv_warning, warning))

                    model.add(regulator)

        return model

    def write(self):

        pass

    def close(self):

        if hasattr(self.io, 'close'):
            self.io.close()

    def clean(self):
        self._dto = None



from multiprocessing import Process, Manager
from collections import OrderedDict
from mewpy.simulation.simulation import SimulationResult, SimulationInterface
from mewpy.model.kinetic import ODEModel
from mewpy.solvers import KineticConfigurations, SolverConfigurations, ODEStatus, ode_solver_instance
import warnings
import numpy as np

def kinetic_solve(model, y0, time_steps, parameters=None, factors=None):

    rates = OrderedDict()
    f = model.get_ode(r_dict=rates, params=parameters, factors=factors)
    solver = ode_solver_instance(f, KineticConfigurations.SOLVER_METHOD)

    C, t, y = solver.solve(y0, time_steps)

    for c in C:
        if c < -1 * SolverConfigurations.RELATIVE_TOL:
            return ODEStatus.ERROR, {}, {}

    # values bellow solver precision will be set to 0
    rates.update({k: 0 for k, v in rates.items() if
                  v < SolverConfigurations.ABSOLUTE_TOL and v > - SolverConfigurations.ABSOLUTE_TOL})
    conc = OrderedDict(zip(model.metabolites.keys(), C))
    return ODEStatus.OPTIMAL, rates, conc, t, y


class KineticThread(Process):
    """
    Solves the ODE inside a thread enabling to impose a timeout limit with thread.join(timeout)
    """

    def __init__(self, model, initial_concentrations=None, time_steps=None, parameters=None, factors=None):
        Process.__init__(self, daemon=False)
        self.model = model
        self.parameters = parameters
        self.factors = factors
        self.initial_concentrations = initial_concentrations
        self.time_steps = time_steps

        self.result = Manager().dict()
        self.result['status'] = None
        self.result["rates"] = None
        self.result["concentrations"] = None
        self.result["t"] = None
        self.result["y"] = None

    def run(self):
        try:
            status, rates, concentrations, t, y = kinetic_solve(self.model,
                                                                self.initial_concentrations,
                                                                self.time_steps,
                                                                self.parameters,
                                                                self.factors)
            self.result['status'] = status
            self.result["rates"] = rates
            self.result["concentrations"] = concentrations
            self.result["t"] = t
            self.result["y"] = y
        except Exception:
            warnings.warn('Timeout')
        return


class KineticSimulationResult(SimulationResult):

    def __init__(self, model, status, factors=None, rates=None, concentrations=None, t=None, y=None):
        super(KineticSimulationResult, self).__init__(model, None, fluxes=rates, status=status)
        self.factors = factors
        self.concentrations = concentrations
        self.t = t
        self.y = y
        self.m_indexes = {k: v for v, k in enumerate(concentrations.keys())} 

    def get_y(self, m_id):
        if m_id in self.m_indexes:
            return np.array(self.y).T[:,self.m_indexes[m_id]]
        else:
            raise ValueError(f"Unknown metabolite {m_id}")

    def get_ss_concentrations(self, format=None):
        if format and format=='df':
            import pandas as pd
            return pd.DataFrame(self.concentrations)
        else:
            return self.concentrations

    def plot(self, met=None, size:tuple=None):    
        import matplotlib.pyplot as plt
        if size:
            plt.rcParams["figure.figsize"] = size
        if not met:
            _mets = list(self.concentrations.keys()) 
        elif isinstance(met,str):
            _mets =[met]
        elif isinstance(met,list) and len(met)<=4:
            _mets = met
        else:
            raise ValueError('fluxes should be a reaction identifier, a list of reaction identifiers or None.')
        ax = plt.subplot()
        if len(_mets)!=2: 
            for k in _mets:
                ax.plot(self.t, self.get_y(k), label=k)
            if len(_mets)==1:
                ax.set_ylabel(self.model.get_metabolite(_mets[0]).name)
            else:        
                ax.set_ylabel('Concentrations')
                plt.legend()
        else:
            ax.plot(self.t, self.get_y(_mets[0]), label=_mets[0])
            ax2 = plt.twinx(ax)
            ax2.plot(self.t, self.get_y(_mets[1]), label=_mets[1], color='r')
            ax.set_ylabel(self.model.get_metabolite(_mets[0]).name, color='b')
            ax2.set_ylabel(self.model.get_metabolite(_mets[1]).name, color='r')
        
        ax.set_xlabel('Time')
        return ax

class KineticSimulation(SimulationInterface):

    def __init__(self, model,  parameters=None, t_points=[0, 1e9], 
                 timeout=KineticConfigurations.SOLVER_TIMEOUT):
        if not isinstance(model, ODEModel):
            raise ValueError('model is not an instance of ODEModel.')
        self.model = model
        self.t_points = t_points
        self.timeout = timeout
        self.parameters = parameters if parameters else dict()

    def get_initial_concentrations(self, initcon=None):
        values = []
        _initcon = initcon if initcon else dict()
        for i, m in enumerate(self.model.metabolites):
            try:
                values.append(_initcon.get(m, self.model.concentrations[m]))
            except :
                values.append(None)
        return values

    def set_time(self, start, end, steps):
        """
        This function sets the time parameters for the model.  This is how long the model will simulate

        Args:
            start (int): the start time - usually 0
            end (int): the end time (default is 100)
            steps (int): the number of timepoints for the output
        """
        self.t_points = np.linspace(start, end, steps)


    def get_time_points(self):
        """Returns the time point or span."""
        return self.t_points

    def simulate(self, parameters=None, initcon=None, factors=None, t_points=None):
        """
        Solve an initial value problem for a system of ODEs.
        
        :param dict parameters: Parameters to be modified. Default None
        :para dict initcon: initial conditions, metabolite concentrations. Default None
        :param dict factors: Modification over the kinetic model.
        :param list t_points: Times at which to store the computed solution,\
            must be sorted and lie within t_span. Default None, in which case the number of
            time steps is defined by SolverConfigurations.N_STEPS.
        :returns: Returns a kineticSimulationResult with the steady-state flux distribution and concentrations.
        """

        _factors = factors if factors is not None else {}
        initConcentrations = self.get_initial_concentrations(initcon)

        status = None
        sstateRates = None
        sstateConc = None
        t = None
        y = None
        params = self.parameters
        if parameters:
            params.update(parameters)
        
        time_steps = t_points if t_points else self.get_time_points()
        
        if len(time_steps)==2:
            time_steps = np.linspace(time_steps[0], 
                                     time_steps[1], 
                                     num=SolverConfigurations.N_STEPS, 
                                     endpoint=True)

        if self.timeout:
            try:
                th = KineticThread(self.model,
                                   initial_concentrations=initConcentrations,
                                   time_steps=time_steps,
                                   parameters=params,
                                   factors=_factors)

                th.start()
                th.join(self.timeout)
                status = th.result['status']
                sstateRates = th.result['rates']
                sstateConc = th.result['concentrations']
                t = th.result['t']
                y = th.result['y']
            except AssertionError as e:
                raise AssertionError(f"{str(e)}. Installing ray for multiprocessing will solve this issue.")     
            except Exception as e:
                warnings.warn(str(e))
        else:
            status, sstateRates, sstateConc, t, y = kinetic_solve(self.model,
                                                            initConcentrations,
                                                            time_steps,
                                                            params,
                                                            _factors)

        return KineticSimulationResult(self.model, status, factors=_factors, rates=sstateRates,
                                       concentrations=sstateConc, t=t, y=y)



import copy
import logging
from typing import List, TypeVar

import numpy
from mewpy.visualization.plot import StreamingPlot
from ..ea import non_dominated_population, Solution

S = TypeVar('S')
LOGGER = logging.getLogger('mewpy')


class VisualizerObserver():

    def __init__(self,
                 reference_front: List[S] = None,
                 reference_point: list = None,
                 display_frequency: float = 1.0,
                 non_dominated=False,
                 axis_labels=None,
                 nevaluations=None) -> None:
        self.figure = None
        self.display_frequency = display_frequency
        self.reference_point = reference_point
        self.reference_front = reference_front
        self.non_dominated = non_dominated
        self.axis_labels = axis_labels
        self.nevaluations = nevaluations

    def update(self, *args, **kwargs):
        evaluations = kwargs['EVALUATIONS']
        solutions = kwargs['SOLUTIONS']
        obj_directions = kwargs['PROBLEM'].obj_directions

        if solutions:
            population = [Solution(s.variables, s.objectives) for s in solutions]

            # check if reference point has changed
            # reference_point = kwargs.get('REFERENCE_POINT', None)
            # negative fitness values are converted to positive
            for i in range(len(population)):
                obj = population[i].fitness
                population[i].fitness = [(-1*obj[k]*obj_directions[k]) for k in range(len(obj))]

            nds = non_dominated_population(population)
            ds = None

            if not self.non_dominated:
                ds = list(set(population)-set(nds))

            if self.figure is None:
                self.figure = StreamingPlot(axis_labels=self.axis_labels)
                self.figure.plot(nds, dominated=ds)
            else:
                text = str(evaluations)
                self.figure.update(nds, dominated=ds, text=text)

            self.figure.ax.set_title(
                'Eval: {}'.format(evaluations), fontsize=13)


class PrintObjectivesStatObserver():

    def __init__(self, frequency: float = 1.0) -> None:
        """ Show the number of evaluations, best fitness and computing time.

        :param frequency: Display frequency. """
        self.display_frequency = frequency
        self.first = True

    def fitness_statistics(self, solutions, problem):
        """Return the basic statistics of the population's fitness values.
        :param list solutions: List of solutions.
        :param problem: The jMetalPy problem.
        :returns: A statistics dictionary.
        """
        def minuszero(value):
            return round(value, 6)

        stats = {}
        first = solutions[0].objectives
        # number of objectives
        n = len(first)
        for i in range(n):
            direction = problem.obj_directions[i]
            factor = 1 if direction == problem.MINIMIZE else -1
            f = [(factor * p.objectives[i]) for p in solutions]
            if direction == problem.MAXIMIZE:
                worst_fit = min(f)
                best_fit = max(f)
            else:
                worst_fit = max(f)
                best_fit = min(f)
            med_fit = numpy.median(f)
            avg_fit = numpy.mean(f)
            std_fit = numpy.std(f)
            stats['obj_{}'.format(i)] = {'best': minuszero(best_fit), 'worst': minuszero(worst_fit),
                                         'mean': minuszero(avg_fit), 'median': minuszero(med_fit), 'std': minuszero(std_fit)}
        return stats

    def stats_to_str(self, stats, evaluations, title=False):
        if title:
            title = "Eval(s)|"
        values = " {0:>6}|".format(evaluations)

        for key in stats:
            s = stats[key]
            if title:
                title = title + "     Worst      Best    Median   Average   Std Dev|"
            values = values + "  {0:.6f}  {1:.6f}  {2:.6f}  {3:.6f}  {4:.6f}|".format(s['worst'],
                                                                                      s['best'],
                                                                                      s['median'],
                                                                                      s['mean'],
                                                                                      s['std'])
        if title:
            return title + "\n" + values
        else:
            return values

    def update(self, *args, **kwargs):
        evaluations = kwargs['EVALUATIONS']
        solutions = kwargs['SOLUTIONS']
        problem = kwargs['PROBLEM']
        if (evaluations % self.display_frequency) == 0 and solutions:
            if type(solutions) == list:
                stats = self.fitness_statistics(solutions, problem)
                message = self.stats_to_str(stats, evaluations, self.first)
                self.first = False
            else:
                fitness = solutions.objectives
                res = abs(fitness[0])
                message = 'Evaluations: {}\tFitness: {}'.format(
                    evaluations, res)
            print(message)



import ast
import os
from collections import OrderedDict
import numpy as np
from .constants import ModelConstants
from ..optimization.ea import Solution, non_dominated_population
from ..optimization.ea import filter_duplicates
from ..simulation import get_simulator, SimulationMethod

# xml escapes
escapes = {'&': '&#38', '<': '&lt', '>': '&gt', "'": '&#39', '"': '&#34'}


def str_constraints(constraints, separator=";"):
    """ Converts a dictionary o constraints into a string representation.
        Constraints may of the form  'key:value' or 'key:(value,value)'
    """
    tokens = []
    for k, v in constraints.items():
        tokens.append(k)
        if isinstance(v, (float, int)):
            tokens.extend([str(v), ""])
        elif isinstance(v, tuple):
            tokens.extend(map(str, list(v)))
        else:
            raise ValueError("Unrecognized value type")
    out_str = separator.join(x for x in tokens)
    return out_str


def xml_constraints(constraints, separator=""):
    """ Converts a dictionary o constraints into a xml string representation.
    Constraints may of the form  'key:value' or 'key:(value,value)'


    :param constrainst: dic, a dictionary of constraints
    :param separator: str, optional
    :return: a xml string representation
    """
    tokens = []
    tokens.append("<constraints>")
    for k, v in constraints.items():
        tokens.append("<constraint><reaction>{}</reaction>".format(k))
        if isinstance(v, (float, int)):
            lb = v
            ub = v
        elif isinstance(v, tuple):
            lb = v[0]
            ub = v[1]
        else:
            raise ValueError("Unrecognized value type")
        tokens.append("<lb>{}</lb>".format(lb))
        tokens.append("<ub>{}</ub>".format(ub))
        tokens.append("</constraint>")
    tokens.append("</constraints>")
    out_str = separator.join(x for x in tokens)
    return out_str


def population_to_csv(problem, candidates, filename, simplify=False, non_dominated=False, sep=';'):
    """ Saves a population of solution candidates to csv
    """
    if non_dominated:
        population = non_dominated_population(
            candidates, problem.is_maximization)
    else:
        population = candidates

    # should probably simplify recursevly until no changes
    if simplify:
        pop = []
        for s in population:
            pop.extend(problem.simplify(s))
        population = pop

    population.sort(reverse=True)
    with open(filename, 'w') as f:
        title = [str(x).replace(sep, " ") for x in problem.fevaluation]
        title.extend([sep, "solution", "\n"])
        f.write(sep.join(x for x in title))
        for individual in population:
            line = []
            if problem.number_of_objectives == 1:
                line.append(str(individual.fitness))
            else:
                line.extend(map(str, individual.fitness))
            line.append("")
            line.append(str(individual.values))
            line.append("")
            line.append(str(individual.constraints))
            line.append("\n")
            f.write(sep.join(x for x in line))


def population_to_xml(problem, candidates, filename, simplify=False, non_dominated=True):
    """ Saves a population of solution candidates to xml
    """

    if non_dominated:
        population = non_dominated_population(
            candidates, problem.is_maximization)
    else:
        population = candidates

    # should recursevly simplify until the initial
    # and simplified populations are the same
    if simplify:
        pop = []
        for s in population:
            pop.extend(problem.simplify(s))
        population = pop

    population.sort(reverse=True)
    with open(filename, 'w') as f:
        f.write("<population>")
        f.write("<ea><objectives>")
        for obj in problem.fevaluation:
            f.write("<objective>{}</objective>".format(str(obj)))
        f.write("</objectives>")
        f.write("</ea><candidates>")
        for individual in population:
            line = []
            line.append("<candidate>")
            if problem.number_of_objectives == 1:
                line.append("<fitness>{}</fitness>".format(individual.fitness))
            else:
                for value in individual.fitness:
                    line.append("<fitness>{}</fitness>".format(value))
            constraints = individual.get_constraints()
            line.append(xml_constraints(constraints))
            line.append("</candidate>")
            f.write("".join(x for x in line))
        f.write("</candidates></population>")


def is_number(s):
    """Easy way to verify if a string is a number
    """
    return s.replace('.', '').replace('-', '').replace('e', '').isdigit()


class Parser:
    def __init__(self, obj_labels):
        """
        Parser to retreive solutions saved as csv.

        :param obj_labels: (list) optimization objective labels.
        """
        self.obj_labels = obj_labels
        self.n_obj = len(obj_labels)
        self.obj_labels.append('Size')
        self.population = []
        self.targets_stats = None

    def __parse_line(self, line, separator, parse_constraints=True):
        line = line.replace(';;', ';')
        tokens = line.split(separator)
        tokens.remove('\n')
        # fitness
        fitness = [abs(float(x)) for x in tokens[:self.n_obj]]
        # encoding
        values = ast.literal_eval(tokens[self.n_obj])

        if parse_constraints:
            # temporary compatibility with previous version
            if tokens[self.n_obj + 1].startswith('{'):
                constrainst = ast.literal_eval(tokens[self.n_obj + 1])
            elif tokens[self.n_obj + 1].startswith('OrderedDict'):
                constrainst = eval(tokens[self.n_obj + 1])
            else:
                constrainst = {}
                i = self.n_obj + 1
                while i < len(tokens) - 2:
                    if is_number(tokens[i + 2]):
                        constrainst[tokens[i]] = (float(tokens[i + 1]), float(tokens[i + 2]))
                        i = i + 3
                    else:
                        constrainst[tokens[i]] = (float(tokens[i + 1]), float(tokens[i + 1]))
                        i = i + 2

        else:
            constrainst = None
        if np.prod(fitness) != 0:
            solution = Solution(values, fitness, constrainst)
            self.population.append(solution)

    def parse_results(self, path, separator=';', non_dominated=True, maximize=True, filter_duplicate=True,
                      parse_constraints=True):
        """
        Parse all csv file in the path.

        Parameters

        :param path: (str) the path to the directory containing the csv files;
        :param separator: (str) the separator used in the csv files;
        :param non_dominated: (boolean) filter dominated solutions;
        :param maximize: (boolean) if the solutions are from a maximization problem;
        :param filter_duplicate: (boolean) remove duplicates solutions.

        """
        for r, _, fl in os.walk(path):
            for file in fl:
                if '.csv' in file:
                    # print('Parsing :',file)
                    with open(os.path.join(r, file), 'r') as f:
                        # skip the first line
                        f.readline()
                        line = f.readline()
                        while line:
                            self.__parse_line(line, separator, parse_constraints=parse_constraints)
                            line = f.readline()
        if non_dominated:
            self.population = non_dominated_population(
                self.population, maximize=maximize, filter_duplicate=filter_duplicate)
        elif filter_duplicate:
            self.population = filter_duplicates(self.population)

    def compute_fluxes(self, model, biomass, product, carbon_source, envcond=None,
                       population=None, minimal_growth=0.0, minimal_product=0.0):
        """
        Computes flux analises for each solution contained in the population.

        :param model: The model.
        :param biomass: (str) Biomass reaction id.
        :param product: (str) Product reaction id.
        :param carbon_source: (str) Carbon source reaction id.
        :param envcond: (dict) Dictionary of environmental conditions. Default None.
        :param population: (list) List of solutions. If none is given the parsed population is considered.
        :param minimal_growth: (float) Solution with growth below "minimal_growth" are discarded. Default 0.0.
        :param minimal_product: (float) Solution with product yield below "minimal_product" are discarded. Default 0.0.

        """
        ModelConstants.RESET_SOLVER = True
        simul = get_simulator(model, envcond=envcond)
        simul.set_objective(biomass)
        reactions = [product, carbon_source]
        if population is None:
            population = self.population
        if not population or len(population) == 0:
            raise RuntimeError("Population is empty")
        res = simul.simulate(
            objective={biomass: 1}, method=SimulationMethod.pFBA)
        self.wt_biomass = res.fluxes[biomass]
        self.wt_products = {}
        # wt_ref = res.fluxes

        for rx in reactions:
            self.wt_products[rx] = res.fluxes[rx]

        filtered_population = []
        data_fitness = []
        # data_constraints = []
        data_fba = []
        data_lmoma = []
        data_fva = []

        n = len(population[0].fitness)
        if n == self.obj_labels:
            labels = self.obj_labels
        else:
            labels = [f"Obj_{i}" for i in range(1, n + 1)]

        for solution in population:
            r = []
            u = []
            m = []

            constraints = {}
            constraints.update(solution.constraints)

            try:
                res = simul.simulate(
                    objective={biomass: 1}, method=SimulationMethod.pFBA, constraints=constraints)
                if res.fluxes:
                    # pFBA
                    biomass_value = res.fluxes[biomass]
                    if biomass_value < minimal_growth or res.fluxes[product] < minimal_product:
                        continue
                    r.append(biomass_value)
                    for rx in reactions:
                        r.append(res.fluxes[rx])
                    # lMOMA
                    res = simul.simulate(objective={
                        biomass: 1}, method=SimulationMethod.lMOMA, constraints=constraints)
                    u.append(res.fluxes[biomass])
                    for rx in reactions:
                        u.append(res.fluxes[rx])
                    # FVA
                    constraints[biomass] = (biomass_value * 0.99, 100000.0)
                    res = simul.simulate(objective={product: 1}, constraints=constraints, maximize=False)
                    if res.fluxes:
                        m.append(res.fluxes[product])
                    else:
                        m.append(0)
                    res = simul.simulate(objective={product: 1}, constraints=constraints, maximize=True)
                    if res.fluxes:
                        m.append(res.fluxes[product])
                    else:
                        m.append(0)

                filtered_population.append(solution)
                data_fitness.append(solution.get_fitness())
                # data_constraints.append(str(solution.constraints))
                data_fba.append(r)
                data_lmoma.append(u)
                data_fva.append(m)
            except Exception:
                continue

        data_values = [str(individual.values) for individual in filtered_population]
        import pandas as pd
        df_values = pd.DataFrame(data_values, columns=["Solution"])
        df_fitness = pd.DataFrame(data_fitness, columns=labels)
        df_fba = pd.DataFrame(data_fba, columns=['Biomass_pFBA', product + '_pFBA', carbon_source + '_pFBA'])
        df_lmoma = pd.DataFrame(data_lmoma, columns=['Biomass_lMOMA', product + '_lMOMA', carbon_source + '_lMOMA'])
        df_fva = pd.DataFrame(data_fva, columns=[product + '_FVAmin_99', product + '_FVAmax_99'])
        # df_constraints = pd.DataFrame(data_constraints, columns=['Constraints'])

        df_all = df_values.join(df_fitness).join(df_fba).join(df_lmoma).join(df_fva)  # .join(df_constraints)

        return df_all, filtered_population

    def reaction_stats(self):
        """
        Reactions statistics
        """
        stats = OrderedDict()
        for solution in self.population:
            rxns = solution.constraints.keys()
            for rx in rxns:
                if rx in stats.keys():
                    stats[rx] += 1
                else:
                    stats[rx] = 1
        self.stats = stats

    def target_stats(self, population=None):
        """
        Target statistics
        """
        stats = OrderedDict()
        if not population:
            population = self.population
        for solution in population:
            targets = list(solution.values.keys())
            for t in targets:
                if t in stats.keys():
                    stats[t] += 1
                else:
                    stats[t] = 1
        import operator
        sorted_d = OrderedDict(sorted(stats.items(),
                                      key=operator.itemgetter(1), reverse=True))

        self.targets_stats = sorted_d
        return sorted_d

    def find_all_solution(self, target, population=None):
        """
        Finds all solutions containing the list of targets

        :param targets: (list) List of target ids.
        :param population: (list) List of solution where to perform the search. If none is provided, the search is \
            performed on the entire population.
        :returns: A list of solutions.
        """
        result = []
        if not isinstance(target, list):
            target = [target]
        if not population:
            population = self.population
        for solution in population:
            if all([v in solution.values.keys() for v in target]):
                result.append(solution)
        return result



from re import findall
from typing import Any, Dict, Generator, Union, TYPE_CHECKING

from mewpy.util.utilities import generator, chemical_formula_re
from mewpy.germ.models.serialization import serialize
from mewpy.util.history import recorder
from mewpy.util.constants import atomic_weights
from .variable import Variable

if TYPE_CHECKING:
    from .reaction import Reaction


class Metabolite(Variable, variable_type='metabolite', register=True, constructor=True, checker=True):

    def __init__(self,
                 identifier: Any,
                 charge: int = None,
                 compartment: str = None,
                 formula: str = None,
                 reactions: Dict[str, 'Reaction'] = None,
                 **kwargs):

        """
        A metabolite is regularly associated with reactions.
        In metabolic-regulatory models, metabolites can be associated with regulators too.

        It holds information regarding the charge, compartment, formula and reactions to which is associated
        Some dynamic information is inferred from the formula such as molecular weight and atoms.
        Other information is inferred from the reactions associated, such as the exchange reaction

        :param identifier: identifier, e.g. h2o_e
        :param charge: the charge of the metabolite
        :param compartment: the compartment of this respective metabolite
        :param formula: a string-like representation of the chemical formula
        :param reactions: the dictionary of reactions to which the metabolite is associated with
        """

        if not charge and charge != 0:
            charge = None

        if not compartment:
            compartment = None

        if not formula:
            formula = ''

        if not reactions:
            reactions = {}

        self._charge = charge
        self._compartment = compartment
        self._formula = formula
        self._reactions = reactions

        super().__init__(identifier,
                         **kwargs)

    # -----------------------------------------------------------------------------
    # Variable type manager
    # -----------------------------------------------------------------------------

    @property
    def types(self):

        # noinspection PyUnresolvedReferences
        _types = {Metabolite.variable_type}

        _types.update(super(Metabolite, self).types)

        return _types

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------
    def __str__(self):

        return f'{self.id} || {self.name} || {self.formula}'

    def _metabolite_to_html(self):
        """
        It returns a html dict representation.
        """
        html_dict = {'Compartment': self.compartment,
                     'Formula': self.formula,
                     'Molecular weight': self.molecular_weight,
                     'Charge': self.charge,
                     'Reactions': ', '.join(self.reactions)}
        return html_dict

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------
    @serialize('charge', 'charge', '_charge')
    @property
    def charge(self) -> int:
        """
        The charge of the metabolite
        :return: charge as int
        """

        if self._charge is None:
            return 0

        return self._charge

    @serialize('compartment', 'compartment', '_compartment')
    @property
    def compartment(self) -> str:
        """
        The compartment of the metabolite
        :return: compartment as str
        """
        return self._compartment

    @serialize('formula', 'formula', '_formula')
    @property
    def formula(self) -> str:
        """
        The chemical formula of the metabolite
        :return: formula as str
        """
        return self._formula

    @serialize('reactions', 'reactions', '_reactions')
    @property
    def reactions(self) -> Dict[str, 'Reaction']:
        """
        The reactions to which the metabolite is associated with
        :return: reactions as dict
        """
        return self._reactions.copy()

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------
    @charge.setter
    @recorder
    def charge(self, value):
        """
        Sets the charge of the metabolite
        :param value: charge as int
        :return:
        """

        if not value and value != 0:
            value = None

        self._charge = value

    @formula.setter
    @recorder
    def formula(self, value):
        """
        Sets the chemical formula of the metabolite
        :param value: formula as str
        :return:
        """

        if not value:
            value = ''

        self._formula = value

    @compartment.setter
    @recorder
    def compartment(self, value):
        """
        Sets the compartment of the metabolite
        :param value: compartment as str
        :return:
        """

        if not value:
            value = None

        self._compartment = value

    @reactions.setter
    def reactions(self, value):
        """
        Sets the reactions to which the metabolite is associated with
        It does not perform additional operations in the associated reactions and models
        :param value: reactions as dict
        :return:
        """

        if not value:
            value = {}

        self._reactions = value

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------
    @property
    def atoms(self) -> Dict[str, int]:
        """
        Atoms and frequency of the metabolite
        :return: atoms as dict
        """

        all_elements = findall(chemical_formula_re, self.formula)

        atoms = {}
        for atom, count in all_elements:

            if not count:
                count = '1'

            atoms[atom] = atoms.get(atom, 0) + int(count)

        return atoms

    @property
    def molecular_weight(self) -> Union[float, int]:
        """
        The molecular weight of the metabolite
        :return: molecular weight as float
        """

        return sum([atomic_weights[atom] * count for atom, count in self.atoms.items()])

    @property
    def exchange_reaction(self) -> 'Reaction':
        """
        The exchange reaction of the metabolite.
        It finds the first boundary reaction in which the metabolite is involved
        :return: exchange reaction as Reaction
        """

        for reaction in self.yield_reactions():
            if reaction.boundary:
                return reaction

    @property
    def exchange_reactions(self) -> Dict[str, 'Reaction']:
        """
        The exchange reactions of the metabolite.
        It finds all boundary reactions in which the metabolite is involved
        :return: exchange reactions as dict
        """

        exchanges = {}

        for reaction in self.yield_reactions():
            if reaction.boundary:
                exchanges[reaction.id] = reaction

        return exchanges

    # -----------------------------------------------------------------------------
    # Generators
    # -----------------------------------------------------------------------------
    def yield_reactions(self) -> Generator['Reaction', None, None]:
        """
        Yields the reactions to which the metabolite is associated with
        :return: reactions as generator
        """

        return generator(self._reactions)

    def yield_exchange_reactions(self) -> Generator['Reaction', None, None]:
        """
        Yields the exchange reactions to which the metabolite is associated with
        :return: exchange reactions as generator
        """
        return generator(self.exchange_reactions)

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------
    def update(self,
               charge: int = None,
               compartment: str = None,
               formula: str = None,
               reactions: Dict[str, 'Reaction'] = None,
               **kwargs):

        """
        It updates the metabolite with the provided information

        Note that, some update operations are not registered in history.
        It is strongly advisable to use update outside history context manager

        :param charge: the charge of the metabolite
        :param compartment: the compartment of this respective metabolite
        :param formula: a string-like representation of the chemical formula
        :param reactions: the dictionary of reactions to which the metabolite is associated with
        :param kwargs: additional attributes
        :return:
        """

        super(Metabolite, self).update(**kwargs)

        if charge is not None:
            self.charge = charge

        if compartment is not None:
            self.compartment = compartment

        if formula is not None:
            self.formula = formula

        if reactions is not None:
            self._reactions.update(reactions)



from typing import TYPE_CHECKING, Union, Dict, Sequence, List, Tuple

import numpy as np
import pandas as pd

from mewpy.germ.analysis import FBA
from mewpy.germ.analysis.analysis_utils import biomass_yield_to_rate, \
    CoRegMetabolite, CoRegBiomass, metabolites_constraints, gene_state_constraints, system_state_update, \
    build_metabolites, build_biomass, CoRegResult
from mewpy.germ.solution import ModelSolution, DynamicSolution
from mewpy.germ.variables import Gene, Target
from mewpy.solvers.solution import Solution, Status
from mewpy.solvers.solver import Solver
from mewpy.util.constants import ModelConstants

if TYPE_CHECKING:
    from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel


def _run_and_decode(lp, additional_constraints=None, solver_kwargs=None):
    if not solver_kwargs:
        solver_kwargs = {}

    if additional_constraints:
        solver_kwargs['constraints'] = {**solver_kwargs.get('constraints', {}), **additional_constraints}

    solution = lp.solver.solve(**solver_kwargs)

    if not solution.values:
        return {rxn: 0 for rxn in lp.model.reactions}, 0

    return solution.values, solution.fobj


def result_to_solution(result: CoRegResult, model: 'Model', to_solver: bool = False) -> Union[ModelSolution, Solution]:
    """
    It converts a CoRegResult object to a ModelSolution object.

    :param result: the CoRegResult object
    :param model: the model
    :param to_solver: if True, it returns a Solution object
    :return: the ModelSolution object
    """
    if to_solver:
        return Solution(status=Status.OPTIMAL, fobj=result.objective_value, values=result.values)

    solution = ModelSolution(method='CoRegFlux',
                             x=result.values,
                             objective_value=result.objective_value,
                             status='optimal',
                             model=model)

    solution.metabolites = {key: met.concentration for key, met in result.metabolites.items()}
    solution.biomass = result.biomass.biomass_yield
    solution.constraints = result.constraints
    return solution


class CoRegFlux(FBA):

    def __init__(self,
                 model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                 solver: Union[str, Solver] = None,
                 build: bool = False,
                 attach: bool = False):
        """
        CoRegFlux is aimed at integrating reverse engineered transcriptional regulatory networks and gene-expression
        into metabolic models to improve prediction of phenotypes.
        It builds a linear regression estimator to predict the expression of target genes
        as function of the co-expression of regulators. The influence score of the regulators is used as input
        for the linear regression model.
        Then, it uses the predicted expression of the target genes to constrain the bounds of the associated reactions.
        It infers continuous constraints on the fluxes of the metabolic model.

        Author: Pauline Trébulle, Daniel Trejo-Banos, Mohamed Elati
        For more detail consult: https://dx.doi.org/10.1186%2Fs12918-017-0507-0
        :param model: a GERM model aka an integrated RegulatoryMetabolicModel
        :param solver: A Solver, CplexSolver, GurobiSolver or OptLangSolver instance.
        Alternatively, the name of the solver is also accepted.
        The solver interface will be used to load and solve a linear problem in a given solver.
        If none, a new solver is instantiated. An instantiated solver may be used,
        but it will be overwritten if build is true.
        :param build: Whether to build the linear problem upon instantiation. Default: False
        :param attach: Whether to attach the linear problem to the model upon instantiation. Default: False
        """
        super().__init__(model=model, solver=solver, build=build, attach=attach)

    # ---------------------------------
    # Dynamic simulation
    # ---------------------------------
    def next_state(self,
                   solver_kwargs: Dict = None,
                   state: Dict[str, float] = None,
                   metabolites: Dict[str, CoRegMetabolite] = None,
                   biomass: CoRegBiomass = None,
                   time_step: float = None,
                   soft_plus: float = 0,
                   tolerance: float = ModelConstants.TOLERANCE,
                   scale: bool = False) -> CoRegResult:
        """
        It computes the next state of the system given the current state and the time step.
        :param solver_kwargs: solver arguments
        :param state: current state of the system
        :param metabolites: metabolites constraints
        :param biomass: biomass constraints
        :param time_step: time step
        :param soft_plus: soft plus parameter
        :param tolerance: tolerance
        :param scale: whether to scale the metabolites
        :return: next state of the system
        """
        # Similar to the Simulation_step in the R implementation
        result = CoRegResult()

        constraints = {reaction.id: reaction.bounds for reaction in self.model.yield_reactions()}

        if metabolites:
            # updating coregflux constraints using metabolites concentrations
            constraints = metabolites_constraints(constraints=constraints,
                                                  metabolites=metabolites,
                                                  biomass=biomass,
                                                  time_step=time_step)

        if state:
            # updating coregflux bounds using gene state (predicted with predict_gene_state from gene expression and
            # regulator coregnet influence score)
            constraints = gene_state_constraints(model=self.model,
                                                 constraints=constraints,
                                                 state=state,
                                                 soft_plus=soft_plus,
                                                 tolerance=tolerance,
                                                 scale=scale)

        # retrieve the fba simulation from the inferred constraints
        values, objective_value = _run_and_decode(self, additional_constraints=constraints, solver_kwargs=solver_kwargs)
        result.values = values
        result.objective_value = objective_value

        # Updating the system state by solving an euler step for metabolites and
        # biomass given the previously fba solution, namely the flux state
        next_biomass, next_metabolites = system_state_update(model=self.model,
                                                             flux_state=values,
                                                             metabolites=metabolites,
                                                             biomass=biomass,
                                                             time_step=time_step,
                                                             biomass_fn=biomass_yield_to_rate)

        result.metabolites = next_metabolites
        result.biomass = next_biomass
        result.constraints = constraints
        return result

    def _dynamic_optimize(self,
                          to_solver: bool = False,
                          solver_kwargs: Dict = None,
                          initial_state: Sequence[Dict[str, float]] = None,
                          metabolites: Dict[str, CoRegMetabolite] = None,
                          biomass: CoRegBiomass = None,
                          time_steps: Sequence[float] = None,
                          soft_plus: float = 0,
                          tolerance: float = ModelConstants.TOLERANCE,
                          scale: bool = False) -> Union[DynamicSolution, Dict[float, Solution]]:
        solutions = []

        previous_time_step = 0
        for i_initial_state, time_step in zip(initial_state, time_steps):
            time_step_diff = time_step - previous_time_step

            next_state = self.next_state(solver_kwargs=solver_kwargs,
                                         state=i_initial_state,
                                         metabolites=metabolites,
                                         biomass=biomass,
                                         time_step=time_step_diff,
                                         soft_plus=soft_plus,
                                         tolerance=tolerance,
                                         scale=scale)

            previous_time_step = time_step

            metabolites = next_state.metabolites
            biomass = next_state.biomass

            solution = result_to_solution(result=next_state, model=self.model, to_solver=to_solver)
            solutions.append(solution)

        if to_solver:
            return dict(zip(time_steps, solutions))

        return DynamicSolution(*solutions, time=time_steps)

    def _steady_state_optimize(self,
                               to_solver: bool = False,
                               solver_kwargs: Dict = None,
                               initial_state: Dict[str, float] = None,
                               metabolites: Dict[str, CoRegMetabolite] = None,
                               biomass: CoRegBiomass = None,
                               time_step: float = 1,
                               soft_plus: float = 0,
                               tolerance: float = ModelConstants.TOLERANCE,
                               scale: bool = False) -> Union[ModelSolution, Solution]:

        result = self.next_state(solver_kwargs=solver_kwargs,
                                 state=initial_state,
                                 metabolites=metabolites,
                                 biomass=biomass,
                                 time_step=time_step,
                                 soft_plus=soft_plus,
                                 tolerance=tolerance,
                                 scale=scale)
        return result_to_solution(result=result, model=self.model, to_solver=to_solver)

    def _optimize(self,
                  to_solver: bool = False,
                  solver_kwargs: Dict = None,
                  initial_state: Union[Dict[str, float], Sequence[Dict[str, float]]] = None,
                  metabolites: Dict[str, CoRegMetabolite] = None,
                  biomass: CoRegBiomass = None,
                  time_steps: Sequence[float] = None,
                  soft_plus: float = 0,
                  tolerance: float = ModelConstants.TOLERANCE,
                  scale: bool = False) -> Union[DynamicSolution, ModelSolution, Solution, List[Solution]]:
        """
        CoRegFlux optimization method.
        It supports steady state and dynamic optimization.
        :param to_solver: Whether to return the solution as a SolverSolution instance. Default: False
        :param solver_kwargs: Keyword arguments to pass to the solver. See LinearProblem.optimize for details.
        :param initial_state: a dictionary of targets ids and expression predictions to set as initial state.
        :param metabolites: a dictionary of metabolites ids and concentrations to set as initial state
        :param biomass: a float value to set as initial biomass
        :param time_steps: a list of time points to simulate the model over
        :param soft_plus: the soft plus parameter to use for the gene state update
        :param tolerance: the tolerance to use for the gene state update
        :param scale: whether to scale the gene state update
        :return: a ModelSolution instance if dynamic is False,
        a DynamicSolution instance otherwise (if to_solver is False)
        """
        if len(initial_state) == 1:
            return self._steady_state_optimize(to_solver=to_solver,
                                               solver_kwargs=solver_kwargs,
                                               initial_state=initial_state[0],
                                               metabolites=metabolites,
                                               biomass=biomass,
                                               time_step=time_steps[0],
                                               soft_plus=soft_plus,
                                               tolerance=tolerance,
                                               scale=scale)

        return self._dynamic_optimize(to_solver=to_solver,
                                      solver_kwargs=solver_kwargs,
                                      initial_state=initial_state,
                                      metabolites=metabolites,
                                      biomass=biomass,
                                      time_steps=time_steps,
                                      soft_plus=soft_plus,
                                      tolerance=tolerance,
                                      scale=scale)

    def optimize(self,
                 to_solver: bool = False,
                 solver_kwargs: Dict = None,
                 initial_state: Union[Dict[str, float], Sequence[Dict[str, float]]] = None,
                 metabolites: Dict[str, float] = None,
                 growth_rate: float = None,
                 time_steps: Sequence[float] = None,
                 soft_plus: float = 0,
                 tolerance: float = ModelConstants.TOLERANCE,
                 scale: bool = False) -> Union[DynamicSolution, ModelSolution, Solution, List[Solution]]:
        """
        CoRegFlux optimization method.
        It supports steady state and dynamic optimization.
        :param to_solver: Whether to return the solution as a SolverSolution instance. Default: False
        :param solver_kwargs: Keyword arguments to pass to the solver. See LinearProblem.optimize for details.
        :param initial_state: a dictionary of targets ids and expression predictions to set as initial state. For dynamic
        optimization, a list of such dictionaries can be provided to set the initial state at each time point.
        :param metabolites: a dictionary of metabolites ids and concentrations to set as initial state
        :param growth_rate: the initial growth rate to set as initial state
        :param time_steps: a list of time points to simulate the model over
        :param soft_plus: the soft plus parameter to use for the gene state update
        :param tolerance: the tolerance to use for the gene state update
        :param scale: whether to scale the gene state update
        :return: a ModelSolution instance if dynamic is False,
        a DynamicSolution instance otherwise (if to_solver is False)
        """
        if not solver_kwargs:
            solver_kwargs = {}
        solver_kwargs['get_values'] = True

        if time_steps is None:
            time_steps = [1]

        if not initial_state:
            initial_state = [{}]

        else:
            if isinstance(initial_state, dict):
                initial_state = [initial_state]

        if len(initial_state) != len(time_steps):
            raise ValueError("The number of time steps must match the number of initial states.")

        if not metabolites:
            metabolites = {}

        if growth_rate is None:
            _, growth_rate = _run_and_decode(self, solver_kwargs=solver_kwargs)

        metabolites = build_metabolites(self.model, metabolites)
        biomass = build_biomass(self.model, growth_rate)
        return self._optimize(to_solver=to_solver,
                              solver_kwargs=solver_kwargs,
                              initial_state=initial_state,
                              metabolites=metabolites,
                              biomass=biomass,
                              time_steps=time_steps,
                              soft_plus=soft_plus,
                              tolerance=tolerance,
                              scale=scale)


# ----------------------------------------------------------------------------------------------------------------------
# Preprocessing using LinearRegression to predict genes expression from regulators co-expression
# Useful for CoRegFlux method
# ----------------------------------------------------------------------------------------------------------------------
def _get_target_regulators(gene: Union['Gene', 'Target'] = None) -> List[str]:
    """
    It returns the list of regulators of a target gene
    :param gene: Target gene
    :return: List of regulators of the target gene
    """
    if gene is None:
        return []

    if gene.is_target():
        return [regulator.id for regulator in gene.yield_regulators()]

    return []


def _filter_influence_and_expression(interactions: Dict[str, List[str]],
                                     influence: pd.DataFrame,
                                     expression: pd.DataFrame,
                                     experiments: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    It filters influence, expression and experiments matrices to keep only the targets and their regulators.
    :param interactions: Dictionary with the interactions between targets and regulators
    :param influence: Influence matrix
    :param expression: Expression matrix
    :param experiments: Experiments matrix
    :return: Filtered influence matrix, filtered expression matrix, filtered experiments matrix
    """
    targets = pd.Index(set(interactions.keys()))
    regulators = pd.Index(set([regulator for regulators in interactions.values() for regulator in regulators]))

    # filter the expression matrix for the target genes only
    expression = expression.loc[expression.index.intersection(targets)].copy()
    influence = influence.loc[influence.index.intersection(regulators)].copy()
    experiments = experiments.loc[experiments.index.intersection(regulators)].copy()
    return influence, expression, experiments


def _predict_experiment(interactions: Dict[str, List[str]],
                        influence: pd.DataFrame,
                        expression: pd.DataFrame,
                        experiment: pd.Series) -> pd.Series:
    try:
        # noinspection PyPackageRequirements
        from sklearn.linear_model import LinearRegression
    except ImportError:
        raise ImportError('The package sklearn is not installed. '
                          'To compute the probability of target-regulator interactions, please install sklearn '
                          '(pip install sklearn).')

    predictions = {}

    for target, regulators in interactions.items():

        if not regulators:
            predictions[target] = np.nan
            continue

        if target not in expression.index:
            predictions[target] = np.nan
            continue

        if not set(regulators).issubset(influence.index):
            predictions[target] = np.nan
            continue

        if not set(regulators).issubset(experiment.index):
            predictions[target] = np.nan
            continue

        # a linear regression model is trained for
        # y = expression of the target gene for all samples
        # x1 = influence score of the regulator 1 in the train data set
        # x2 = influence score of the regulator 2 in the train data set
        # x3 ...

        x = influence.loc[regulators].transpose().to_numpy()
        y = expression.loc[target].to_numpy()

        regressor = LinearRegression()
        regressor.fit(x, y)

        # the expression of the target gene is predicted for the experiment
        x_pred = experiment.loc[regulators].to_frame().transpose().to_numpy()
        predictions[target] = regressor.predict(x_pred)[0]

    return pd.Series(predictions)


def predict_gene_expression(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                            influence: pd.DataFrame,
                            expression: pd.DataFrame,
                            experiments: pd.DataFrame) -> pd.DataFrame:
    """
    It predicts the expression of genes in the experiments set using the co-expression of regulators
    in the expression and influence datasets.
    Adapted from CoRegFlux docs:
        - A GEM model containing GPRs
        - A TRN network containing target genes and the co-activators and co-repressors of each target gene
        - GEM genes and TRN targets must match
        - An influence dataset containing the influence scores of the regulators. Influence score is similar to a
        correlation score between the expression of the regulator and the expression of the target gene.
        Influence dataset format: (rows: regulators, columns: samples). Influence scores are calculated using the
        CoRegNet algorithm in the gene expression dataset.
        - A gene expression dataset containing the expression of the genes. Also called the training dataset.
        The gene expression dataset format: (rows: genes, columns: samples)
        - An experiments dataset containing the influence scores of the regulators in the experiments. These experiments
        are not used for training the linear regression model (the gene state predictor).
        These experiments are condition-specific environmental or genetic conditions that can be used
        to perform phenotype simulations.
        Experiments dataset format: (rows: regulators, columns: experiments/conditions)

    The result is a matrix of predicted gene expression values for each experiment.
    :param model: an integrated Metabolic-Regulatory model aka a GERM model
    :param influence: Influence scores of the regulators in the train data set
    :param expression: Expression of the genes in the train data set
    :param experiments: Influence scores of the regulators in the test data set
    :return: Predicted expression of the genes in the test data set
    """
    # Filtering only the gene expression and influences data of metabolic genes available in the model
    interactions = {target.id: _get_target_regulators(target) for target in model.yield_targets()}
    influence, expression, experiments = _filter_influence_and_expression(interactions=interactions,
                                                                          influence=influence,
                                                                          expression=expression,
                                                                          experiments=experiments)

    predictions = []
    for column in experiments.columns:
        experiment_prediction = _predict_experiment(interactions=interactions,
                                                    influence=influence,
                                                    expression=expression,
                                                    experiment=experiments[column])
        predictions.append(experiment_prediction)

    predictions = pd.concat(predictions, axis=1)
    predictions.columns = experiments.columns
    return predictions.dropna()



from ..util.utilities import Singleton


class EAGlobalSettings(Singleton):
    POPULATION_SIZE = 100


def get_default_population_size():
    return EAGlobalSettings.POPULATION_SIZE


def set_default_population_size(size: int):
    EAGlobalSettings.POPULATION_SIZE = size



from .variable import Variable, build_variable
from .gene import Gene
from .interaction import Interaction
from .metabolite import Metabolite
from .reaction import Reaction
from .regulator import Regulator
from .target import Target



from copy import deepcopy
from warnings import warn
from collections import OrderedDict
from .problem import AbstractKOProblem
from mewpy.model import CommunityModel
from mewpy.simulation import get_simulator
from mewpy.simulation.simulation import Simulator



class CommunityKOProblem(AbstractKOProblem):
    """
    Community Knockout Optimization Problem.

    :param models: A list of metabolic models.
    :param list fevaluation: A list of callable EvaluationFunctions.

    Optional parameters:

    :param OrderedDict envcond: Environmental conditions.
    :param OrderedDict constraints: Additional constraints to be applied to the model.
    :param int candidate_min_size: The candidate minimum size (Default EAConstants.MIN_SOLUTION_SIZE)
    :param int candidate_max_size: The candidate maximum size (Default EAConstants.MAX_SOLUTION_SIZE)
    :param list target: List of modification target reactions.
    :param list non_target: List of non target reactions. Not considered if a target list is provided.
    :param float scalefactor: A scaling factor to be used in the LP formulation.
    
    """

    def __init__(self, models: list, fevaluation=[], copy_models=True, **kwargs):
        super(CommunityKOProblem, self).__init__(
            None, fevaluation=fevaluation, **kwargs)
        self.organisms = OrderedDict()
        self.model_ids = list({model.id for model in models})

        if len(self.model_ids) < len(models):
            warn("Model ids are not unique, repeated models will be discarded.")

        for model in models:

            m = model if isinstance(model, Simulator) else get_simulator(model)
            self.organisms[m.id] = deepcopy(m) if copy_models else m

    def _build_target_list(self):
        """Target organims, i.e., organisms that may be removed from the community.
        """
        print("Building modification target list.")
        target = set(self.model_ids)
        if self.non_target is not None:
            target = target - set(self.non_target)
        self._trg_list = list(target)

    def solution_to_constraints(self, candidate):
        """Returns a community model that includes all non
        KO organisms.
        TODO: this is a naif approach. Insteas consider using weights
        to turn organism ON/OFF within the community.

        :param candidate: [description]
        :return: [description]
        """
        ko_organisms = list(candidate.keys())
        models = [x for x in self.model_ids if x not in ko_organisms]
        cmodel = CommunityModel(models)
        return cmodel.get_community_model()

    def evaluate_solution(self, solution, decode=True):
        """
        Evaluates a single solution, a community.

        :param solution: The solution to be evaluated (a community model).
        :param decode: If the solution needs to be decoded (convert a list of model ids to a community model).
        :returns: A list of fitness values.
        """
        decoded = {}
        # decoded constraints
        if decode:
            decoded = self.decode(solution)
            cmodel = self.solution_to_constraints(decoded)
        else:
            cmodel = solution
        try:
            p = []
            simulation_results = dict()
            for method in self.methods:
                sim = get_simulator(cmodel, self.environmental_conditions)
                simulation_result = self.simulator.simulate(method=method, scalefactor=self.scalefactor)
                simulation_results[method] = simulation_result
            # apply the evaluation function(s)
            for f in self.fevaluation:
                v = f(simulation_results, decoded,
                      scalefactor=self.scalefactor)
                p.append(v)
        except Exception as e:
            p = []
            for f in self.fevaluation:
                p.append(f.worst_fitness)
        return p





from typing import Any, Union, Type, TYPE_CHECKING, List, Set, Dict, Iterable, Tuple

from mewpy.util.history import HistoryManager, recorder
from mewpy.germ.models.serialization import serialize, Serializer

# Preventing circular dependencies that only happen due to type checking
if TYPE_CHECKING:
    from mewpy.germ.models import MetabolicModel, RegulatoryModel
    from mewpy.germ.variables import Gene, Interaction, Metabolite, Reaction, Regulator, Target, Variable
    from mewpy.germ.lp import LinearProblem


class MetaModel(type):
    """
    The MetaModel class is used to dynamically create Model classes.
    Models instances are created from static classes representing the different information for a given model.

    However, integrated Metabolic-Regulatory models must combine different types of variables, such as genes, reactions,
    metabolites, but also regulators, targets, interactions. This is not possible with static classes, as they miss
    the containers to store these variables.
    Thus, a Model factory is used to dynamically create Model classes
    based on the multiple types of a single integrated model.

    The Model factory is the interface to create these multi-type models.
    It also manages all models by implementing base attributes and methods.

    The MetaModel class creates a dynamic Model class as follows:
        1. It collects all static base classes to be included into the dynamic Model class
        2. It collects the model type from the base classes or from the model_type attribute
        3. It sets the name of the dynamic Model class using the name of the base classes
        4. It updates the model_type attribute with all types of the base classes
        5. It creates the dynamic Model class having all attributes and methods of the base classes
        6. It adds polymorphic constructors to the dynamic Model class based on the types of the base classes
        7. It adds type checkers to the dynamic Model class based on the types of the base classes
    """
    factories = {}

    def __new__(mcs, name, bases, attrs, **kwargs):
        # if it is the model factory, only registration is done
        factory = kwargs.get('factory', False)

        if factory:
            cls = super(MetaModel, mcs).__new__(mcs, name, bases, attrs)

            MetaModel.factories[name] = cls

            return cls

        # Dynamic typing being used. In this case, a proper name and model type must be provided
        dynamic = kwargs.get('dynamic', False)
        if dynamic:
            names = [base.model_type for base in bases]

            name = ''.join([name.title() for name in names])
            name += 'Model'

            kwargs['model_type'] = '-'.join(names)

        # The model type is always added to the subclasses. If it is not given upon subclass creation,
        # the subclass name is to be used
        model_type = kwargs.get('model_type', name.lower())
        attrs['model_type'] = model_type

        return super(MetaModel, mcs).__new__(mcs, name, bases, attrs)

    def __init__(cls, name, bases, attrs, **kwargs):
        super().__init__(name, bases, attrs)

        factory = kwargs.get('factory', False)
        if factory:
            # collection of all containers that must be serialized for a particular class or subclass
            containers = cls.get_serializable_containers(attrs)

            attrs['_containers_registry']['model'] = containers

            # Skip further building of the Model factory
            return

        # Dynamic typing being used. In this case, all children have already been constructed, so everything can be
        # skipped
        dynamic = kwargs.get('dynamic', False)
        if dynamic:
            return

        # Several attributes and methods must be added automatically to the Model factory children based on the
        # model type.
        model_type = attrs['model_type']

        for base in bases:

            factory = MetaModel.factories.get(base.__name__)

            if factory and hasattr(cls, 'model_type'):

                # if some class inherits from the Model factory, it must be registered in the factory for type
                # checking
                if cls.model_type not in factory.get_registry():
                    raise TypeError(f'{cls.model_type} does not inherit from Model')

                # If set otherwise upon class creation, all subclasses of the Model factory will contribute to
                # their parent factory with an alternative initializer. The polymorphic constructor, e.g. from_{
                # model_type}, is added to the Model factory
                constructor = kwargs.get('constructor', True)

                if constructor:
                    # noinspection PyProtectedMember
                    model_type_initializer = Model._from(cls)
                    setattr(factory, f'from_{model_type}', model_type_initializer)

                # If set otherwise upon class creation, all subclasses of the Model factory will contribute to
                # their parent factory with a declared type checker. The type checker, e.g. is_{model_type},
                # is added to the Model factory
                checker = kwargs.get('checker', True)

                if checker:
                    # noinspection PyProtectedMember
                    model_type_checker = Model._is(model_type)
                    setattr(factory, f'is_{model_type}', model_type_checker)

                # collection of all containers that must be serialized for a particular class or subclass
                containers = cls.get_serializable_containers(attrs)

                Model.register_containers(containers, cls)

    @staticmethod
    def get_serializable_containers(attrs: Dict[str, Any]) -> Dict[str, Tuple[str, str, str]]:
        """
        Collects all containers that must be serialized for a particular class or subclass

         All containers have property descriptors marked with the @serialize decorator. This decorator marks attributes
        for serialization, deserialization and pickle serialization/deserialization. The decorator also provides
        information about the attribute name.

        If serialization, deserialization or pickle serialization/deserialization is not desired, the attribute must
        be marked with the @serialize decorator with the corresponding argument set to None.
        :param attrs: the attributes of the class or subclass
        :return: a dictionary with the container name as key and a tuple with the attribute name for serialization,
        deserialization and pickle serialization/deserialization
        """
        containers = {}

        for name, method in attrs.items():

            if hasattr(method, 'fget'):

                if hasattr(method.fget, 'serialize') and hasattr(method.fget, 'deserialize') and hasattr(method.fget,
                                                                                                         'pickle'):
                    containers[name] = (method.fget.serialize, method.fget.deserialize, method.fget.pickle)

        return containers


class Model(Serializer, metaclass=MetaModel, factory=True):
    """
    The Model class is the base class and factory for all models.
    It is the interface to create multi-type models providing several base attributes and methods.

    It implements the basic attributes:
        - types: the types of the model
        - id: the id of the model
        - name: the name of the model
        - simulators: the simulation methods attached to the model. Linear problems such as FBA, RFBA, etc.
        - compartments: the compartments of the model
        - contexts: the contexts associated with the model.
        - history: the history of the model.

    It implements the basic methods:
        - factory: the factory for dynamic model types
        - from_types: creates a model from a list of types
        - copy: creates a copy of the model
        - deepcopy: creates a deep copy of the model
        - get: returns a variable in the model
        - add: adds a variable to the model
        - remove: removes a variable from the model
        - update: updates the model with a new information
        - attach: attaches a simulation method to the model
        - detach: detaches a simulation method from the model
        - notify: notifies all simulation methods attached to the model regarding a change in the model
        - undo: undoes the last change of the model
        - redo: redoes the last change of the model
        - reset: resets the model to its initial state
        - restore: restores the model to a previous state
        - from polymorphic constructors: creates a model from a specific type
        - is polymorphic type checkers: checks if a model is of a specific type

    It implements the basic serialization methods:
        - to_dict: serializes the model to a dictionary
        - from_dict: deserializes the model from a dictionary

    """
    # -----------------------------------------------------------------------------
    # Factory management
    # -----------------------------------------------------------------------------
    _registry = {}
    _containers_registry = {'model': {}}

    def __init_subclass__(cls, **kwargs):
        """
        This method is called when a subclass of Model is created. It is used to register the subclass in the
        Model factory.

        Internal use only.
        :param kwargs: the keyword arguments
        """
        super(Model, cls).__init_subclass__(**kwargs)

        # the child type
        model_type = getattr(cls, 'model_type', cls.__name__.lower())

        cls.register_type(model_type, cls)

    @staticmethod
    def get_registry() -> Dict[str, Type['Model']]:
        """
        Returns the registry of the Model factory.

        Internal use only.
        :return: the registry of the Model factory
        """
        return Model._registry.copy()

    @staticmethod
    def get_containers_registry() -> Dict[str, Dict[str, Tuple[str, str, str]]]:
        """
        Returns the containers registry of the Model factory.

        Internal use only.
        :return: the containers registry of the Model factory
        """
        return Model._containers_registry.copy()

    @staticmethod
    def register_type(model_type: str, child: Type['Model']):
        """
        Registers a type in the Model factory.

        Internal use only.
        :param model_type: the type to register
        :param child: the child class
        :return:
        """
        Model._registry[model_type] = child

    # -----------------------------------------------------------------------------
    # Serialization
    # -----------------------------------------------------------------------------
    @staticmethod
    def register_containers(containers, child):
        """
        Registers the containers of a child class in the containers registry of the Model factory. This is useful for
        serialization and deserialization.

        Internal use only.
        :param containers: the containers to register
        :param child: the child class
        :return:
        """
        if hasattr(child, 'model_type'):

            Model._containers_registry[child.model_type] = containers

        elif child is Model:
            Model._containers_registry['model'] = containers

    @property
    def containers(self) -> Dict[str, Tuple[str, str, str]]:
        """
        Returns the containers of the model. This is useful for serialization and deserialization.

        Internal use only.
        :return: the containers of the model as a dictionary with the container name as key and a tuple with the
        attribute name for serialization, deserialization and pickle serialization/deserialization
        """
        class_containers = self.get_containers_registry()

        containers = class_containers['model'].copy()

        for model_type in self.types:

            for name, (serialize_name, deserialize_name, pickle_name) in class_containers[model_type].items():
                containers[name] = (serialize_name, deserialize_name, pickle_name)

        return containers

    # -----------------------------------------------------------------------------
    # Factory polymorphic constructor
    # -----------------------------------------------------------------------------
    @classmethod
    def factory(cls, *args: str) -> Union[Type['Model'],
                                          Type['MetabolicModel'],
                                          Type['RegulatoryModel']]:
        """
        It creates a dynamic Model class from a list of types. The types must be registered in the Model factory.

        Example:
            >>> Model.factory('metabolic', 'regulatory')
            <class 'mewpy.germ.model.MetabolicRegulatoryModel'>

        :param args: the types of the model
        :return: the model type
        """
        if not args:
            args = ()

        registry = cls.get_registry()

        types = tuple([registry[name] for name in args])

        if len(types) == 1:
            return types[0]

        _Model = MetaModel('Model', types, {}, dynamic=True)

        # noinspection PyTypeChecker
        return _Model

    # -----------------------------------------------------------------------------
    # Factory polymorphic initializer
    # -----------------------------------------------------------------------------
    @classmethod
    def from_types(cls, types: Iterable[str], **kwargs) -> Union['Model',
                                                                 'MetabolicModel',
                                                                 'RegulatoryModel']:
        """
        It creates a model instance from a list of types and a dictionary of containers and attributes.
        The types must be registered in the Model factory.

        :param types: the types of the model
        :param kwargs: the containers and attributes to initilize the model
        :return: the model instance
        """
        ModelType = cls.factory(*types)
        return ModelType(**kwargs)

    # -----------------------------------------------------------------------------
    # Factory helper methods
    # -----------------------------------------------------------------------------
    @staticmethod
    def _from(cls):
        """
        Method to be added to the subclasses of Model to create a model from a specific type.

        Internal use only.
        :param cls: the class of the model
        :return: the model to be added to the subclass of Model
        """

        def from_(identifier, **kwargs):
            return cls(identifier, **kwargs)

        return from_

    @staticmethod
    def _is(model_type):
        """
        Method to be added to the subclasses of Model to check if a model is of a specific type.

        Internal use only.
        :param model_type: the type to check
        :return: the method to be added to the subclass of Model
        """

        def is_(self):
            return self.is_a(model_type)

        return is_

    # -----------------------------------------------------------------------------
    # Base initializer
    # -----------------------------------------------------------------------------
    def __init__(self,
                 identifier: Any,
                 name: str = None):

        """
        The model is the base class for all models, such as metabolic model or regulatory model.
        See also model.MetabolicModel and model.RegulatoryModel for concrete implementations of a model type.

        The model object is the factory for several model types.
        This model object provides all batteries to create new model types and to manage the model types' inheritance.

        The factory type assists with systematic tasks such as attributes and containers serialization,
        history management, simulator observation, among others.

        For that, attributes, containers and inheritance are registered in the factory. The MetaModel manages
        both the factory and the model types derived from the factory.

        :param identifier: identifier,  e.g. iMC1010
        :param name: the name of the model
        """
        self._check_inheritance()

        if not identifier:
            identifier = ''

        if not name:
            name = identifier

        self._id = ''
        self._name = ''
        self._simulators = []
        self._types = set()

        # History: reversible changes to the model
        self._contexts = []
        self._history = HistoryManager()

        self._id = identifier
        self.name = name

        # models share different compartments. Although it is not the most subtle solution, name mangling is used to
        # store the specific compartments of each child. An alternative would be to check for shared attributes
        # between all children and perform this solution for them
        self.__compartments = {}

    def _check_inheritance(self):
        """
        It checks if the class is a subclass of Model and if it is not, it raises an error.
        :return:
        """
        registry = self.get_registry()

        for model_type in self.types:
            if model_type not in registry:
                raise ValueError(f'{model_type} is not registered as subclass of {self.__class__.__name__}')

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------

    def __str__(self):
        return f'Model {self.id} - {self.name}'

    def __repr__(self):
        return self.__str__()

    # noinspection PyUnresolvedReferences
    def _repr_html_(self):
        """
        It returns a html representation of the gene.
        """

        objective = getattr(self, 'objective', None)
        if objective:
            objective = next(iter(objective)).id
        else:
            objective = None

        if self.is_metabolic() and self.is_regulatory():
            return f"""
            <table>
                <tr>
                    <th><b>Model</b></th>
                    <td>{self.id}</td>
                </tr>
                <tr>
                    <th>Name</th>
                    <td>{self.name}</td>
                </tr>
                <tr>
                    <th>Types</th>
                    <td>{', '.join(self.types)}</td>
                </tr>
                <tr>
                    <th>Compartments</th>
                    <td>{', '.join(self.compartments)}</td>
                </tr>
                <tr>
                    <th>Reactions</th>
                    <td>{len(self.reactions)}</td>
                </tr>
                <tr>
                    <th>Metabolites</th>
                    <td>{len(self.metabolites)}</td>
                </tr>
                <tr>
                    <th>Genes</th>
                    <td>{len(self.genes)}</td>
                </tr>
                <tr>
                    <th>Exchanges</th>
                    <td>{len(self.exchanges)}</td>
                </tr>
                <tr>
                    <th>Demands</th>
                    <td>{len(self.demands)}</td>
                </tr>
                <tr>
                    <th>Sinks</th>
                    <td>{len(self.sinks)}</td>
                </tr>
                <tr>
                    <th>Objective</th>
                    <td>{objective}</td>
                </tr>
                <tr>
                    <th>Regulatory interactions</th>
                    <td>{len(self.interactions)}</td>
                </tr>
                <tr>
                    <th>Targets</th>
                    <td>{len(self.targets)}</td>
                </tr>
                <tr>
                    <th>Regulators</th>
                    <td>{len(self.regulators)}</td>
                </tr>
                <tr>
                    <th>Regulatory reactions</th>
                    <td>{len(self.regulatory_reactions)}</td>
                </tr>
                <tr>
                    <th>Regulatory metabolites</th>
                    <td>{len(self.regulatory_metabolites)}</td>
                </tr>
                <tr>
                    <th>Environmental stimuli</th>
                    <td>{len(self.environmental_stimuli)}</td>
                </tr>
            </table>
            """
        elif self.is_metabolic():
            return f"""
            <table>
                <tr>
                    <th><b>Model</b></th>
                    <td>{self.id}</td>
                </tr>
                <tr>
                    <th>Name</th>
                    <td>{self.name}</td>
                </tr>
                <tr>
                    <th>Types</th>
                    <td>{', '.join(self.types)}</td>
                </tr>
                <tr>
                    <th>Compartments</th>
                    <td>{', '.join(self.compartments)}</td>
                </tr>
                <tr>
                    <th>Reactions</th>
                    <td>{len(self.reactions)}</td>
                </tr>
                <tr>
                    <th>Metabolites</th>
                    <td>{len(self.metabolites)}</td>
                </tr>
                <tr>
                    <th>Genes</th>
                    <td>{len(self.genes)}</td>
                </tr>
                <tr>
                    <th>Exchanges</th>
                    <td>{len(self.exchanges)}</td>
                </tr>
                <tr>
                    <th>Demands</th>
                    <td>{len(self.demands)}</td>
                </tr>
                <tr>
                    <th>Sinks</th>
                    <td>{len(self.sinks)}</td>
                </tr>
                <tr>
                    <th>Objective</th>
                    <td>{objective}</td>
                </tr>
            </table>
            """
        elif self.is_regulatory():
            return f"""
            <table>
                <tr>
                    <th><b>Model</b></th>
                    <td>{self.id}</td>
                </tr>
                <tr>
                    <th>Name</th>
                    <td>{self.name}</td>
                </tr>
                <tr>
                    <th>Types</th>
                    <td>{', '.join(self.types)}</td>
                </tr>
                <tr>
                    <th>Compartments</th>
                    <td>{', '.join(self.compartments)}</td>
                </tr>
                <tr>
                    <th>Regulatory interactions</th>
                    <td>{len(self.interactions)}</td>
                </tr>
                <tr>
                    <th>Targets</th>
                    <td>{len(self.targets)}</td>
                </tr>
                <tr>
                    <th>Regulators</th>
                    <td>{len(self.regulators)}</td>
                </tr>
                <tr>
                    <th>Regulatory reactions</th>
                    <td>{len(self.regulatory_reactions)}</td>
                </tr>
                <tr>
                    <th>Regulatory metabolites</th>
                    <td>{len(self.regulatory_metabolites)}</td>
                </tr>
                <tr>
                    <th>Environmental stimuli</th>
                    <td>{len(self.environmental_stimuli)}</td>
                </tr>
            </table>
            """
        return f"""
                <table>
                    <tr>
                        <th><b>Model</b></th>
                        <td>{self.id}</td>
                    </tr>
                    <tr>
                        <th>Name</th>
                        <td>{self.name}</td>
                    </tr>
                    <tr>
                        <th>Types</th>
                        <td>{', '.join(self.types)}</td>
                    </tr>
                </table>
                """

    # -----------------------------------------------------------------------------
    # Model type manager
    # -----------------------------------------------------------------------------
    @serialize('types', None, None)
    @property
    def types(self) -> Set[str]:
        """
        It returns the types of the model.
        :return: the types of the model as a set of strings
        """
        return set()

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------
    @serialize('id', None, '_id')
    @property
    def id(self) -> Any:
        """
        It returns the identifier of the model.
        :return: the identifier of the model
        """
        return self._id

    @serialize('name', 'name', '_name')
    @property
    def name(self) -> str:
        """
        It returns the name of the model.
        :return: the name of the model
        """
        return self._name

    @property
    def simulators(self) -> List['LinearProblem']:
        """
        It returns the list of simulation methods associated with the model.
        :return: the list of simulation methods
        """
        return self._simulators

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------
    @name.setter
    @recorder
    def name(self, value: str):
        """
        It sets the name of the model.
        :param value: the name of the model
        :return:
        """
        if not value:
            value = ''

        self._name = value

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------

    def get(self, identifier: Any, default=None) -> Union['Gene',
                                                          'Interaction',
                                                          'Metabolite',
                                                          'Reaction',
                                                          'Regulator',
                                                          'Target']:
        """
        It returns the variable with the given identifier.
        :param identifier: the identifier of the variable
        :param default: the default value to return if the variable is not found
        :return: the variable with the given identifier
        """
        return default

    def add(self,
            *variables: 'Variable',
            comprehensive: bool = True,
            history: bool = True):
        """
        It adds the given variables to the model.
        This method accepts a single variable or a list of variables to be added to specific containers in the model.
        The containers to which the variables will be added are specified by the types.

        For instance, if a variable is simultaneously a metabolite and regulator,
        it will be added to the metabolites and regulators containers.

        If comprehensive is True, the variables and their related variables will be added to the model too.
        If history is True, the changes will be recorded in the history.

        This method notifies all simulators with the recent changes.

        :param variables: the variables to be added to the model
        :param comprehensive: if True, the variables and their related variables will be added to the model too
        :param history: if True, the changes will be recorded in the history
        :return:
        """
        if history:
            self.history.queue_command(undo_func=self.remove,
                                       undo_args=variables,
                                       undo_kwargs={'remove_orphans': True,
                                                    'history': False},
                                       func=self.add,
                                       args=variables,
                                       kwargs={'comprehensive': comprehensive,
                                               'history': history})

        self.notify()

    def remove(self,
               *variables: 'Variable',
               remove_orphans: bool = False,
               history: bool = True):
        """
        It removes the given variables from the model.
        This method accepts a single variable or a list of variables to be removed from specific containers
        in the model.
        The containers from which the variables will be removed are specified by the types.

        For instance, if a variable is simultaneously a metabolite and regulator,
        it will be removed from the metabolites and regulators containers.

        If remove_orphans is True, the variables and their related variables will be removed from the model too.
        If history is True, the changes will be recorded in the history.

        This method notifies all simulators with the recent changes.

        :param variables: the variables to be removed from the model
        :param remove_orphans: if True, the variables and their related variables will be removed from the model too
        :param history: if True, the changes will be recorded in the history
        :return:
        """
        if history:
            self.history.queue_command(undo_func=self.add,
                                       undo_args=variables,
                                       undo_kwargs={'comprehensive': True,
                                                    'history': False},
                                       func=self.remove,
                                       args=variables,
                                       kwargs={'remove_orphans': remove_orphans,
                                               'history': history})

        self.notify()

    def update(self, name: str = None):
        """
        It updates the model with relevant information.
        :param name: the name of the model
        :return:
        """
        if name:
            self.name = name

    def _add_variable_to_container(self, variable, container):
        """
        It adds the given variable to the given container.

        Helper method to be used internally.
        :param variable: the variable to be added to the container
        :param container: the container to which the variable will be added
        :return:
        """

        if isinstance(container, str):
            container = getattr(self, container, container)

        if variable.id not in container:
            container[variable.id] = variable

            if variable.model is not self:
                variable.model = self
                # noinspection PyProtectedMember
                variable._model_ref = 1

            elif variable.model is self:
                # noinspection PyProtectedMember
                variable._model_ref += 1

    def _remove_variable_from_container(self, variable, container):
        """
        It removes the given variable from the given container.

        Helper method to be used internally.
        :param variable: the variable to be removed from the container
        :param container: the container from which the variable will be removed
        :return:
        """
        if isinstance(container, str):
            container = getattr(self, container, container)

        if variable.id in container:

            del container[variable.id]

            # noinspection PyProtectedMember
            if variable._model_ref < 2:
                variable.model = None

            elif variable._model_ref > 1:
                # noinspection PyProtectedMember
                variable._model_ref -= 1

    # -----------------------------------------------------------------------------
    # Simulators observer pattern
    # -----------------------------------------------------------------------------
    def attach(self, simulator: 'LinearProblem'):
        """
        It attaches the given simulation method (simulator) to the model.
        Once a simulator is attached to the model, it will be notified with model changes.
        Hence, the simulator is synchronized with the model and following optimizations reflect
        the changes in the model.

        :param simulator: the simulator to be attached to the model
        :return:
        """
        self.simulators.append(simulator)

    def detach(self, simulator):
        """
        It detaches the given simulation method (simulator) from the model.
        Once a simulator is detached from the model, it will not be notified with model changes.
        Hence, the simulator is not synchronized with the model and following optimizations do not reflect
        the changes in the model.

        :param simulator: the simulator to be detached from the model
        :return:
        """
        self.simulators.remove(simulator)

    def notify(self):
        """
        It notifies all simulators with the recent changes in the model.

        :return:
        """
        for simulator in self.simulators:
            simulator.update()

    # -----------------------------------------------------------------------------
    # History manager command pattern
    # -----------------------------------------------------------------------------
    @property
    def contexts(self) -> List[HistoryManager]:
        """
        It returns the list of contexts the model is currently in.
        :return: the list of contexts the model is currently in
        """
        return self._contexts

    @property
    def history(self) -> HistoryManager:
        """
        It returns the history manager of the model.
        If the model is in a context, the history manager of the last context is returned.
        :return: the history manager of the model
        """
        if self.contexts:
            return self.contexts[-1]

        return self._history

    def undo(self):
        """
        It undoes the last action performed on the model.
        :return:
        """
        self.history.undo()

    def redo(self):
        """
        It redoes the last action performed on the model.
        :return:
        """
        self.history.redo()

    def clean_history(self):
        """
        It cleans the history of the model.
        :return:
        """
        self._history = HistoryManager()

    def clean_context(self, index=None):
        """
        It cleans the context of the model.
        :param index: the index of the context to be cleaned
        :return:
        """
        if not index:
            index = -1

        self.contexts[index] = HistoryManager()

    def reset(self):
        """
        It resets the model to its initial state.
        :return:
        """
        self.history.reset()

    def restore(self):
        """
        It restores the model to its last state.
        :return:
        """
        self.history.restore()

    def __enter__(self):
        """
        It enters the context of the model.
        It creates a new context and adds it to the list of contexts.
        All changes performed within this context will be reset when the context is exited.

        :return: the model itself
        """
        self.contexts.append(HistoryManager())

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        It exits the context of the model performing a history reset.

        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        context = self.contexts.pop()
        context.reset()

    # -----------------------------------------------------------------------------
    # Abstract implementations
    # -----------------------------------------------------------------------------
    # The following polymorphic initializers are just registered here to avoid type checking errors
    @classmethod
    def from_metabolic(cls,
                       identifier: Any,
                       name: str = None,
                       compartments: Dict[str, str] = None,
                       genes: Dict[str, 'Gene'] = None,
                       metabolites: Dict[str, 'Metabolite'] = None,
                       objective: Dict['Reaction', Union[float, int]] = None,
                       reactions: Dict[str, 'Reaction'] = None) -> 'MetabolicModel':
        """
        It creates a metabolic model from the given information.
        :param identifier: the identifier of the model
        :param name: the name of the model
        :param compartments: the compartments of the model
        :param genes: the genes of the model
        :param metabolites: the metabolites of the model
        :param objective: the objective of the model
        :param reactions: the reactions of the model
        :return: a metabolic model
        """
        ...

    @classmethod
    def from_regulatory(cls,
                        identifier: Any,
                        name: str = None,
                        compartments: Dict[str, str] = None,
                        interactions: Dict[str, 'Interaction'] = None,
                        regulators: Dict[str, 'Regulator'] = None,
                        targets: Dict[str, 'Target'] = None) -> 'RegulatoryModel':
        """
        It creates a regulatory model from the given information.
        :param identifier: the identifier of the model
        :param name: the name of the model
        :param compartments: the compartments of the model
        :param interactions: the interactions of the model
        :param regulators: the regulators of the model
        :param targets: the targets of the model
        :return: a regulatory model
        """
        ...

    # -----------------------------------------------------------------------------
    # Type checker
    # -----------------------------------------------------------------------------
    # The following type checkers are just registered here to avoid type checking errors
    def is_metabolic(self) -> bool:
        """
        It checks whether the model is a metabolic model.
        :return: True if the model is a metabolic model, False otherwise
        """
        ...

    def is_regulatory(self) -> bool:
        """
        It checks whether the model is a regulatory model.
        :return: True if the model is a regulatory model, False otherwise
        """
        ...

    def is_a(self, model_type) -> bool:
        """
        It checks whether the model is of the given type.
        :param model_type: the type of the model
        :return: True if the model is of the given type, False otherwise
        """
        if model_type in self.types:
            return True

        return False

    # -----------------------------------------------------------------------------
    # Common attributes - Name mangling
    # -----------------------------------------------------------------------------
    @property
    def compartments(self) -> Dict[str, str]:
        """
        It returns the compartments of the model.
        :return: the compartments of the model as a dictionary
        """
        return self.__compartments

    # -----------------------------------------------------------------------------
    # Helper methods
    # -----------------------------------------------------------------------------

    @staticmethod
    def _get_orphans(to_remove, first_container, second_container):
        """
        It returns the orphans of the given containers.

        Internal use only.
        :param to_remove: the variables to be removed
        :param first_container: the first container
        :param second_container: the second container
        :return:
        """
        orphans = set()

        for variable in to_remove:

            container_iter = getattr(variable,
                                     f'yield_{first_container}',
                                     lambda: [getattr(variable, f'{first_container}')])

            for variable_2 in container_iter():

                remove_variable = True

                container_iter2 = getattr(variable_2,
                                          f'yield_{second_container}',
                                          lambda: [getattr(variable_2, f'{second_container}')])

                for variable_3 in container_iter2():

                    if variable_3 not in to_remove:
                        remove_variable = False

                if remove_variable:
                    orphans.add(variable_2)

        return orphans


def build_model(types: Iterable[str], kwargs: Dict[str, Any]) -> Union['Model', 'MetabolicModel', 'RegulatoryModel']:
    """
    It builds a model from the given types and arguments. Check the `Model.from_types()` method for details.
    :param types: the types of the model
    :param kwargs: the arguments of the model
    :return: a new model instance for the given types and arguments
    """
    return Model.from_types(types, **kwargs)



from mewpy.problems import GOUProblem

class HybridGOUProblem(GOUProblem):

    def __init__(self, model, hconstraints:dict, fevaluation:list=None,**kwargs):
        """ Overrides GOUProblem by applying constraints resulting from
        sampling a kinetic model.

        :param model: The constraint metabolic model.
        :param list fevaluation: A list of callable EvaluationFunctions.
        :param dict hconstraints: The hybrid constraints definind kinetic model solution space. 
        Optional:

        :param OrderedDict envcond: Environmental conditions.
        :param OrderedDict constraints: Additional constraints to be applied to the model.
        :param int candidate_min_size: The candidate minimum size (Default EAConstants.MIN_SOLUTION_SIZE)
        :param int candidate_max_size: The candidate maximum size (Default EAConstants.MAX_SOLUTION_SIZE)
        :param list target: List of modification target genes.
        :param list non_target: List of non target genes. Not considered if a target list is provided.
        :param float scalefactor: A scaling factor to be used in the LP formulation.
        :param dic reference: Dictionary of flux values to be used in the over/under expression values computation.
        :param tuple operators: (and, or) operations. Default (MIN, MAX).
        :param list levels: Over/under expression levels (Default EAConstants.LEVELS).
        :param boolean twostep: If deletions should be applied before identifiying reference flux values.

        """
        super().__init__(model, fevaluation, **kwargs)
        self.hconstraints=hconstraints

    def solution_to_constraints(self, candidate):
        constraints = super().solution_to_constraints(candidate)
        # Apply the hybrid contraints:
        # Finds the intersection ranges of the kinetic and steady-state 
        # constraints genetic modifications space.
        # If not overlapping, the genetic modifications are not applied. 
        for r,v in  self.hconstraints.items():
            if r in constraints.keys():
                x = constraints[r]
                if isinstance(x,tuple):
                    lb,ub = x
                else:
                    lb=x
                    ub=x  
                hlb,hub = v
                l=max(lb,hlb)
                u=min(ub,hub)
                if l<=u:
                    constraints[r]=(l,u)
                else:
                    constraints[r]=(hlb,hub)
            else:
                constraints[r]=v
        return constraints    




import pandas as pd


class Summary:
    """
    A summary of a ModelSolution. This object contains the main information of the solution.
    """
    def __init__(self,
                 inputs: pd.DataFrame = None,
                 outputs: pd.DataFrame = None,
                 objective: pd.DataFrame = None,
                 df: pd.DataFrame = None,
                 metabolic: pd.DataFrame = None,
                 regulatory: pd.DataFrame = None):
        """
        A summary of a ModelSolution

        :param inputs: the inputs of the model
        :param outputs: the outputs of the model
        :param objective: the objective of the model
        :param df: the data frame of the summary
        :param metabolic: the metabolic summary
        :param regulatory: the regulatory summary
        :return:
        """
        if inputs is None:
            inputs = pd.DataFrame()

        if outputs is None:
            outputs = pd.DataFrame()

        if objective is None:
            objective = pd.DataFrame()

        if df is None:
            df = pd.DataFrame()

        if metabolic is None:
            metabolic = pd.DataFrame()

        if regulatory is None:
            regulatory = pd.DataFrame()

        self.inputs = inputs
        self.outputs = outputs
        self.objective = objective
        self.df = df
        self.metabolic = metabolic
        self.regulatory = regulatory

    def _repr_html_(self):
        """
        It returns a html representation of the linear problem
        :return:
        """
        return self.df.to_html()



from typing import Union, TYPE_CHECKING, Dict, Optional, Tuple, Any

import pandas as pd

from .summary import Summary
from mewpy.util.constants import ModelConstants

if TYPE_CHECKING:
    from mewpy.solvers import Solution
    from mewpy.germ.variables import Variable
    from mewpy.germ.lp import LinearProblem
    from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel


class ModelSolution:
    """
    ModelSolution can be used to retrieve the results of a simulation using metabolic, regulatory
    or integrated analysis.
    It contains the main information about the solution:
        - method
        - variables values (x)
        - objective value
        - status
        - objective direction
        - reduced costs
        - shadow prices

    It also contains the model and the simulator used to obtain the solution.

    All solution attributes can be retrieved directly from a ModelSolution object.
    Alternatively, three export methods are available:
        - to_series: returns a pandas Series with the fluxes of the reactions and regulatory variables
        - to_frame: returns a pandas DataFrame with the fluxes of the reactions, regulatory variables
        and environmental conditions
        - to_summary: returns a pandas DataFrame with the input and output fluxes of the reactions, regulatory variables
         and the environmental conditions
    """
    def __init__(self,
                 method: str,
                 x: Dict[str, float],
                 objective_value: float,
                 status: str,
                 objective_direction: str = 'maximize',
                 reduced_costs: Dict[str, float] = None,
                 shadow_prices: Dict[str, float] = None,
                 model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
                 simulator: 'LinearProblem' = None,
                 tol: float = ModelConstants.TOLERANCE):
        """
        ModelSolution can be used to retrieve the results of a simulation using metabolic, regulatory
        or integrated analysis.
        It contains the main information about the solution:
            - method
            - variables values (x)
            - objective value
            - status
            - objective direction
            - reduced costs
            - shadow prices

        It also contains the model and the simulator used to obtain the solution.

        All solution attributes can be retrieved directly from a ModelSolution object.
        Alternatively, three export methods are available:
            - to_series: returns a pandas Series with the fluxes of the reactions and regulatory variables
            - to_frame: returns a pandas DataFrame with the fluxes of the reactions, regulatory variables
            and environmental conditions
            - to_summary: returns a pandas DataFrame with the input and output fluxes of the reactions,
            regulatory variables and the environmental conditions

        :param method: analysis method used to obtain the solution
        :param x: dictionary with the variables values
        :param objective_value: objective value obtained in the simulation
        :param status: the status of the solution obtained from the solver
        :param objective_direction: the direction of the objective function. Default is 'maximize'
        :param reduced_costs: The reduced costs of the variables. Default is None
        :param shadow_prices: The shadow prices of the constraints. Default is None
        :param model: The model used to obtain the solution. Default is None
        :param simulator: The simulator used to obtain the solution. Default is None
        :param tol: The tolerance used to round the solution. Default is 1e-6
        """
        if not method:
            method = 'fba'

        if not x:
            x = {}

        if not objective_value:
            objective_value = 0.0

        if not status:
            status = 'feasible'

        if not objective_direction:
            objective_direction = 'maximize'

        if not reduced_costs:
            reduced_costs = {}

        if not shadow_prices:
            shadow_prices = {}

        self._method = method
        self._x = x
        self._objective_value = objective_value
        self._status = status
        self._objective_direction = objective_direction
        self._reduced_costs = reduced_costs
        self._shadow_prices = shadow_prices
        self._model = model
        self._simulator = simulator
        self.tol = tol

    # ---------------------------------
    # Buil-in
    # ---------------------------------
    def __str__(self):
        return f'{self.method} Solution\n Objective value: {self.objective_value}\n Status: {self.status}'

    def __repr__(self):
        return self.__str__()

    def _repr_html_(self):
        """
        It returns a html representation of the linear problem
        :return:
        """

        return f"""
        <table>
            <tr>
                <td>Method</td>
                <td>{self.method}</td>
            </tr>
            <tr>
                <td>Model</td>
                <td>{self.model}</td>
            </tr>
            <tr>
                <th>Objective</th>
                <td>{self.objective}</td>
            </tr>
            <tr>
                <th>Objective value</th>
                <td>{self.objective_value}</td>
            </tr>
            <tr>
                <th>Status</th>
                <td>{self.status}</td>
            </tr>
        </table>
        """

    @staticmethod
    def _filter_mid_term_variables(x: Dict[str, float],
                                   model: Union['Model', 'MetabolicModel', 'RegulatoryModel']) -> Dict[str, float]:
        """
        It filters out midterm variables from the solution.
        Internal use only.
        :param x: The solution dictionary
        :param model: The model used to obtain the solution
        :return: A dictionary with the solution without mid-term variables
        """
        if model is None:
            return x

        new_x = {}

        for variable, value in x.items():

            model_variable = model.get(variable, None)

            if model_variable is not None:
                new_x[variable] = value

        return new_x

    @property
    def objective_direction(self) -> str:
        """
        The direction of the objective function.
        :return: The direction of the objective function
        """
        return self._objective_direction

    @property
    def method(self) -> str:
        """
        The method used to obtain the solution.
        :return: The method used to obtain the solution
        """
        return self._method

    @property
    def model(self) -> Union['Model', 'MetabolicModel', 'RegulatoryModel']:
        """
        The model used to obtain the solution.
        :return: The model used to obtain the solution
        """
        return self._model

    @property
    def objective(self) -> Optional[str]:
        """
        A string representation of the objective function of the model.
        :return: A string representation of the objective function of the model
        """
        if self.model:

            if hasattr(self.model, 'objective'):
                return ' + '.join([obj.id for obj in self.model.objective])

        return

    @property
    def objective_value(self) -> float:
        """
        The objective value obtained in the simulation.
        :return: The objective value obtained in the simulation
        """
        return self._objective_value

    @property
    def simulator(self) -> 'LinearProblem':
        """
        The simulator used to obtain the solution.
        :return: A LinearProblem-like object defining the simulator used to obtain the solution
        """
        return self._simulator

    @property
    def status(self) -> str:
        """
        The status of the solution obtained from the solver.
        :return: The status of the solution obtained from the solver
        """
        return self._status

    @property
    def x(self) -> Dict[str, float]:
        """
        The variables' values obtained in the solution of the linear problem.
        :return: A dictionary with the variables values
        """
        return self._filter_mid_term_variables(x=self._x, model=self.model)

    @property
    def shadow_prices(self):
        """
        The shadow prices of the variables. How much of each variable would be needed to
        increase the objective value.
        :return: A dictionary with the shadow prices of the constraints
        """
        return self._filter_mid_term_variables(x=self._shadow_prices, model=self.model)

    @property
    def reduced_costs(self):
        """
        The reduced costs of the variables. The objective value to increase
        to assume a positive value in the optimal solution.
        :return: A dictionary with the reduced costs of the variables
        """
        return self._filter_mid_term_variables(x=self._reduced_costs, model=self.model)

    def _get_variable_info(self, variable: 'Variable') -> Tuple[Any, str, Optional[float]]:
        """
        It returns the information about a variable in the solution.
        Internal use only.
        :param variable:
        :return:
        """
        identifier = variable.id

        x = self._x.get(variable.id, None)

        v_type = ', '.join(variable.types)

        return identifier, v_type, x

    def _metabolic_frame(self) -> pd.DataFrame:
        """
        It returns a pandas DataFrame with the fluxes of the reactions and environmental conditions.
        Internal use only.
        :return: A pandas DataFrame with the fluxes of the reactions and environmental conditions
        """

        results = {}

        for variable in self.model.yield_reactions():
            _id, v_type, x = self._get_variable_info(variable)

            results[_id] = (variable.id, v_type, x)

        return pd.DataFrame.from_dict(results,
                                      orient='index',
                                      columns=['reaction', 'variable type', 'flux'])

    def _regulatory_frame(self) -> pd.DataFrame:
        """
        It returns a pandas DataFrame with the regulatory variables.
        Internal use only.
        :return: A pandas DataFrame with the regulatory variables
        """
        results = {}

        for variable in self.model.yield_regulators():
            _id, v_type, x = self._get_variable_info(variable)

            results[_id] = (variable.id, v_type, x)

        for variable in self.model.yield_targets():
            _id, v_type, x = self._get_variable_info(variable)

            results[_id] = (variable.id, v_type, x)

        return pd.DataFrame.from_dict(results,
                                      orient='index',
                                      columns=['regulatory variable',
                                               'variable type',
                                               'expression coefficient'])

    def _metabolic_environmental_conditions_frame(self) -> pd.DataFrame:
        """
        It returns a pandas DataFrame with the environmental conditions.
        Internal use only.
        :return: A pandas DataFrame with the environmental conditions
        """
        results = {}

        for variable in self.model.yield_exchanges():

            if variable.is_reaction():
                _id, v_type, x = self._get_variable_info(variable)

                metabolite = next(iter(variable.metabolites.keys()))

                lb, ub = variable.bounds

                results[_id] = (_id, v_type, metabolite, lb, ub, x)

        return pd.DataFrame.from_dict(results,
                                      orient='index',
                                      columns=['exchange', 'variable type', 'metabolite',
                                               'lower bound', 'upper bound',
                                               'flux'])

    def _regulatory_environmental_conditions_frame(self) -> pd.DataFrame:
        """
        It returns a pandas DataFrame with the environmental conditions.
        Internal use only.
        :return: A pandas DataFrame with the environmental conditions
        """
        results = {}

        for variable in self.model.yield_environmental_stimuli():

            if variable.is_regulator():
                _id, v_type, x = self._get_variable_info(variable)

                lb, ub = variable.coefficients

                results[_id] = (_id, v_type, lb, ub, x)

        return pd.DataFrame.from_dict(results,
                                      orient='index',
                                      columns=['regulatory variable', 'variable type',
                                               'minimum coefficient', 'maximum coefficient',
                                               'expression coefficient'])

    def _regulatory_inputs_frame(self) -> pd.DataFrame:
        """
        It returns a pandas DataFrame with the inputs of the regulatory model.
        Internal use only.
        :return: A pandas DataFrame with the inputs of the regulatory model
        """
        results = {}

        for variable in self.model.yield_environmental_stimuli():

            _id, v_type, x = self._get_variable_info(variable)

            if variable.is_regulator():
                results[_id] = (_id, v_type, x)

        return pd.DataFrame.from_dict(results,
                                      orient='index',
                                      columns=['regulator', 'variable type', 'expression coefficient'])

    def _regulatory_outputs_frame(self) -> pd.DataFrame:
        """
        It returns a pandas DataFrame with the outputs of the regulatory model.
        Internal use only.
        :return: A pandas DataFrame with the outputs of the regulatory model
        """
        results = {}

        for variable in self.model.yield_targets():
            _id, v_type, x = self._get_variable_info(variable)

            results[_id] = (_id, v_type, x)

        return pd.DataFrame.from_dict(results,
                                      orient='index',
                                      columns=['target', 'variable type', 'expression coefficient'])

    def _metabolic_inputs_frame(self) -> pd.DataFrame:
        """
        It returns a pandas DataFrame with the inputs of the metabolic model.
        Internal use only.
        :return: A pandas DataFrame with the inputs of the metabolic model
        """
        results = {}

        for variable in self.model.yield_exchanges():

            _id, v_type, x = self._get_variable_info(variable)

            if variable.is_reaction() and x < -self.tol:
                metabolite = next(iter(variable.metabolites.keys()))

                results[_id] = (_id, v_type, metabolite, x)

        return pd.DataFrame.from_dict(results,
                                      orient='index',
                                      columns=['reaction', 'variable type', 'metabolite', 'flux'])

    def _metabolic_outputs_frame(self) -> pd.DataFrame:
        """
        It returns a pandas DataFrame with the outputs of the metabolic model.
        Internal use only.
        :return: A pandas DataFrame with the outputs of the metabolic model
        """
        results = {}

        for variable in self.model.yield_exchanges():

            _id, v_type, x = self._get_variable_info(variable)

            if variable.is_reaction() and x > self.tol:
                metabolite = next(iter(variable.metabolites.keys()))

                results[_id] = (_id, v_type, metabolite, x)

        return pd.DataFrame.from_dict(results,
                                      orient='index',
                                      columns=['reaction', 'variable type', 'metabolite', 'flux'])

    def _metabolic_summary_frame(self) -> pd.DataFrame:
        """
        It returns a pandas DataFrame with the summary of the metabolic model.
        Internal use only.
        :return: A pandas DataFrame with the summary of the metabolic model
        """
        results = {}

        for variable in self.model.yield_exchanges():

            _id, v_type, x = self._get_variable_info(variable)

            if variable.is_reaction() and x < -self.tol:
                metabolite = next(iter(variable.metabolites.keys()))

                results[_id] = (_id, v_type, metabolite, 'input', x)

            if variable.is_reaction() and x > self.tol:
                metabolite = next(iter(variable.metabolites.keys()))

                results[_id] = (_id, v_type, metabolite, 'output', x)

        return pd.DataFrame.from_dict(results,
                                      orient='index',
                                      columns=['reaction', 'variable type', 'metabolite', 'role', 'flux'])

    def _regulatory_summary_frame(self) -> pd.DataFrame:
        """
        It returns a pandas DataFrame with the summary of the regulatory model.
        Internal use only.
        :return: A pandas DataFrame with the summary of the regulatory model
        """
        results = {}

        for variable in self.model.yield_environmental_stimuli():

            _id, v_type, x = self._get_variable_info(variable)

            if variable.is_regulator():
                results[_id] = (_id, v_type, 'input', x)

        for variable in self.model.yield_targets():
            _id, v_type, x = self._get_variable_info(variable)

            results[_id] = (_id, v_type, 'output', x)

        return pd.DataFrame.from_dict(results,
                                      orient='index',
                                      columns=['regulatory variable', 'variable type', 'role',
                                               'expression coefficient'])

    def _objective_frame(self) -> pd.DataFrame:
        """
        It returns a pandas DataFrame with the objective value.
        Internal use only.
        :return: A pandas DataFrame with the objective value
        """
        return pd.DataFrame([[self.objective_value, self.objective_direction]],
                            index=[self.objective],
                            columns=['value', 'direction'])

    def _get_frame(self, to: str, dimension: str) -> pd.DataFrame:
        """
        It returns a pandas DataFrame with the results of the model.
        Internal use only.
        :param to: The type of the results to be returned
        :param dimension: The dimension of the results to be returned
        :return: A pandas DataFrame with the results of the model
        """
        dim_to = f'{dimension}_{to}'

        if dim_to == 'regulatory_environmental_conditions':

            return self._regulatory_environmental_conditions_frame()

        elif dim_to == 'metabolic_environmental_conditions':

            return self._metabolic_environmental_conditions_frame()

        elif dim_to == 'regulatory_frame':

            return self._regulatory_frame()

        elif dim_to == 'metabolic_frame':

            return self._metabolic_frame()

        elif dim_to == 'metabolic_inputs':

            return self._metabolic_inputs_frame()

        elif dim_to == 'metabolic_outputs':

            return self._metabolic_outputs_frame()

        elif dim_to == 'regulatory_inputs':

            return self._regulatory_inputs_frame()

        elif dim_to == 'regulatory_outputs':

            return self._regulatory_outputs_frame()

        elif dim_to == 'objective_objective':

            return self._objective_frame()

        elif dim_to == 'regulatory_summary':

            return self._regulatory_summary_frame()

        elif dim_to == 'metabolic_summary':

            return self._metabolic_summary_frame()

        else:
            return pd.DataFrame()

    def to_frame(self, dimensions: Tuple[str, ...] = None) -> pd.DataFrame:
        """
        It returns a  pandas DataFrame having the summary results of the simulation.

        Example:
        >>> from mewpy.germ.analysis import FBA
        >>> from mewpy.io import read_sbml
        >>> model = read_sbml('e_coli_core.xml')
        >>> fba = FBA(model)
        >>> solution = fba.optimize()
        >>> solution.to_frame()

        :param dimensions: The dimensions of the results to be returned. If None, all dimensions are returned.
        possible values are: 'regulatory', 'metabolic', 'objective'
        :return: A pandas DataFrame having the summary results of the model
        """
        if not dimensions:
            dimensions = self.model.types

        frames = [self._get_frame(to='environmental_conditions', dimension=dimension)
                  for dimension in dimensions]

        if frames:
            return pd.concat(frames,
                             axis=1,
                             join='outer',
                             keys=dimensions)

        return pd.DataFrame()

    def to_series(self) -> pd.Series:
        """
        It returns a pandas Series with the values of the linear problem variables.
        Namely, the X variables can include reaction fluxes, metabolite concentrations, gene variables,
        and regulatory variables.
        :return: A pandas Series with the values of the linear problem variables
        """
        return pd.Series(self.x)

    def _summary_builder(self, dimensions: Tuple[str, ...]) -> Summary:
        """
        It returns a namedtuple with pandas DataFrame having all results of the model.
        Internal use only.
        :param dimensions: The dimensions of the results to be returned
        :return: A Summary with pandas DataFrame having all results of the model
        """
        frames = {}

        inputs_frames = [self._get_frame(to='inputs', dimension=dimension)
                         for dimension in dimensions]
        if inputs_frames:
            inputs_frame = pd.concat(inputs_frames,
                                     axis=1,
                                     join='outer',
                                     keys=dimensions)
        else:
            inputs_frame = pd.DataFrame()
        frames['inputs'] = inputs_frame

        outputs_frames = [self._get_frame(to='outputs', dimension=dimension)
                          for dimension in dimensions]
        if outputs_frames:
            outputs_frame = pd.concat(outputs_frames,
                                      axis=1,
                                      join='outer',
                                      keys=dimensions)
        else:
            outputs_frame = pd.DataFrame()
        frames['outputs'] = outputs_frame

        objective_frame = self._get_frame(to='objective', dimension='objective')
        frames['objective'] = objective_frame

        summary_frames = [self._get_frame(to='summary', dimension=dimension)
                          for dimension in dimensions]
        if summary_frames:
            frame = pd.concat(summary_frames,
                              axis=1,
                              join='outer',
                              keys=dimensions)

        else:
            frame = pd.DataFrame()
        frames['df'] = frame

        for i, dimension in enumerate(dimensions):
            frames[dimension] = summary_frames[i]

        return Summary(**frames)

    def to_summary(self, dimensions: Tuple[str, ...] = None) -> Summary:
        """
        It returns a summary with pandas DataFrame of the simulation.

        According to the dimensions, the Summary will have the following attributes:
            - inputs: A pandas DataFrame with the inputs of the metabolic and regulatory model
            - outputs: A pandas DataFrame with the outputs of the metabolic and regulatory model
            - objective: A pandas DataFrame with the objective value
            - metabolic: A pandas DataFrame with the summary of the metabolic model
            - regulatory: A pandas DataFrame with the summary of the regulatory model
            - df: A pandas DataFrame with the summary of the metabolic and regulatory models

        Example:
        >>> from mewpy.germ.analysis import FBA
        >>> from mewpy.io import read_sbml
        >>> model = read_sbml('e_coli_core.xml')
        >>> fba = FBA(model)
        >>> solution = fba.optimize()
        >>> solution.to_summary()

        :param dimensions: The dimensions of the results to be returned. If None, all dimensions are returned.
        possible values are: 'regulatory', 'metabolic', 'objective'
        :return: A namedtuple with pandas DataFrame having all results of the model
        """
        if not dimensions:
            dimensions = self.model.types

        return self._summary_builder(dimensions=dimensions)

    @classmethod
    def from_solver(cls,
                    method: str,
                    solution: 'Solution',
                    **kwargs) -> 'ModelSolution':
        """
        It returns a solution object from a solver solution.
        :param method: The method used to solve the problem
        :param solution: The solution object returned by the solver
        :param kwargs: Additional arguments
        :return: A new ModelSolution object
        """
        minimize = kwargs.pop('minimize', False)
        if minimize:
            objective_direction = 'minimize'
        else:
            objective_direction = 'maximize'

        return cls(method=method,
                   x=solution.values,
                   objective_value=solution.fobj,
                   objective_direction=objective_direction,
                   status=solution.status.value.lower(),
                   reduced_costs=solution.reduced_costs,
                   shadow_prices=solution.shadow_prices,
                   **kwargs)



from setuptools import setup, find_packages
import sys

files = ["model/data/*"]

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = ['cobra', 'inspyred', 'jmetalpy',
                'reframed', 'networkx', 'matplotlib',
                'joblib', 'tdqm']

setup_requirements = requirements + ['pytest-runner']
test_requirements = requirements + ['pytest', 'cplex']
install_requirements = requirements

setup(
    name='mewpy',
    version='0.1.21',
    python_requires='>=3.6',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    package_data={"": ["*.xml", "*.csv", "*.txt"], 'mewpy': files},
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requirements,
    setup_requires=setup_requirements,
    tests_require=test_requirements,
    author='BiSBII CEB University of Minho',
    author_email='vpereira@ceb.uminho.pt',
    description='mewpy - Metabolic Engineering in Python ',
    license='Apache License Version 2.0',
    keywords='strain optimization',
    url='https://github.com/BioSystemsUM/mewpy/',
    long_description=readme,
    test_suite='tests',
)



from ..solvers.solution import Status, print_values
from ..solvers import solver_instance
from ..util.utilities import molecular_weight
from warnings import warn
from math import inf, isinf


def SteadyCom(community, constraints=None, solver=None):
    """ Implementation of SteadyCom (Chan et al 2017). Adapted from REFRAMED
    Args:
        community (CommunityModel): community model
        constraints (dict): environmental or additional constraints (optional)
        solver (Solver): solver instance instantiated with the model, for speed (optional)
    Returns:
        CommunitySolution: solution object
    """

    if solver is None:
        solver = build_problem(community)

    objective = community.comm_model.objective
    sol = binary_search(solver, objective, minimize=False, constraints=constraints)

    solution = CommunitySolution(community, sol.values)
    solution.solver = solver
    return solution


def SteadyComVA(community, obj_frac=1.0, constraints=None, solver=None):
    """ Abundance Variability Analysis using SteadyCom (Chan et al 2017). Adapated from REFRAMED
    Args:
        community (CommunityModel): community model
        obj_frac (float): minimum fraction of the maximum growth rate (default 1.0)
        constraints (dict): environmental or additional constraints (optional)
        solver (Solver): solver instance instantiated with the model, for speed (optional)
    Returns:
        dict: species abundance variability
    """

    if solver is None:
        solver = build_problem(community)

    objective = community.comm_model.objective

    sol = binary_search(solver, objective, constraints=constraints)
    growth = obj_frac * sol.values[community.biomass]
    solver.update_growth(growth)

    variability = {org_id: [None, None] for org_id in community.organisms}

    for org_id in community.organisms:
        sol2 = solver.solve({f"x_{org_id}": 1}, minimize=True, get_values=False, constraints=constraints)
        variability[org_id][0] = sol2.fobj

    for org_id in community.organisms:
        sol2 = solver.solve({f"x_{org_id}": 1}, minimize=False, get_values=False, constraints=constraints)
        variability[org_id][1] = sol2.fobj

    return variability


def build_problem(community, growth=1, bigM=1000):

    solver = solver_instance()
    model = community.get_community_model()

    # create biomass variables
    for org_id in community.organisms.keys():
        solver.add_variable(f"x_{org_id}", 0, 1, update=False)

    # create all community reactions
    for r_id in model.reactions:
        reaction = model.get_reaction(r_id)
        if r_id in model.get_exchange_reactions():
            solver.add_variable(r_id, reaction.lb, reaction.ub, update=False)
        else:
            lb = -inf if reaction.lb < 0 else 0
            ub = inf if reaction.ub > 0 else 0
            solver.add_variable(r_id, lb, ub, update=False)

    solver.update()

    # sum biomass = 1
    solver.add_constraint("abundance", {f"x_{org_id}": 1 for org_id in community.organisms.keys()},
                          rhs=1, update=False)

    # S.v = 0
    table = model.metabolite_reaction_lookup()
    for m_id in model.metabolites:
        solver.add_constraint(m_id, table[m_id], update=False)

    # organism-specific constraints
    for org_id, organism in community.organisms.items():

        for r_id in organism.reactions:
            if (org_id, r_id) not in community.reaction_map:
                continue

            new_id = community.reaction_map[(org_id, r_id)]

            # growth = mu * X
            if r_id in organism.objective.keys():
                solver.add_constraint(f"g_{org_id}", {f"x_{org_id}": growth, new_id: -1}, update=False)
            # lb * X < R < ub * X
            else:
                lb = -bigM if isinf(reaction.lb) else reaction.lb
                ub = bigM if isinf(reaction.ub) else reaction.ub

                if lb != 0:
                    solver.add_constraint(f"lb_{new_id}", {f"x_{org_id}": lb, new_id: -1}, '<', 0, update=False)

                if ub != 0:
                    solver.add_constraint(f"ub_{new_id}", {f"x_{org_id}": ub, new_id: -1}, '>', 0, update=False)

    solver.update()

    def update_growth(value):
        coefficients = [(f"g_{x}", f"x_{x}", value) for x in community.organisms]
        solver.change_coefficients(coefficients)

    solver.update_growth = update_growth

    return solver


def binary_search(solver, objective, obj_frac=1, minimize=False, max_iters=30, abs_tol=1e-3, constraints=None):
    previous_value = 0
    value = 1
    fold = 2
    feasible = False
    last_feasible = 0

    for i in range(max_iters):
        diff = value - previous_value

        if diff < abs_tol:
            break

        if feasible:
            last_feasible = value
            previous_value = value
            value = fold*diff + value
        else:
            if i > 0:
                fold = 0.5
            value = fold*diff + previous_value

        solver.update_growth(value)
        sol = solver.solve(objective, get_values=False, minimize=minimize, constraints=constraints)

        feasible = sol.status == Status.OPTIMAL

    if feasible:
        solver.update_growth(obj_frac * value)
    else:
        solver.update_growth(obj_frac * last_feasible)
    sol = solver.solve(objective, minimize=minimize, constraints=constraints)

    if i == max_iters - 1:
        warn("Max iterations exceeded.")

    return sol


class CommunitySolution(object):

    def __init__(self, community, values):
        self.community = community
        self.values = values
        self.growth = None
        self.abundance = None
        self.exchange = None
        self.internal = None
        self.normalized = None
        self._exchange_map = None
        self.parse_values()

    def __str__(self):
        growth = f"Community growth: {self.growth}\n"
        abundance = "\n".join(f"{org_id}\t{val}" for org_id, val in self.abundance.items())
        return growth + abundance

    def __repr__(self):
        return str(self)

    @property
    def exchange_map(self):
        if self._exchange_map is None:
            self._exchange_map = self.compute_exchanges()

        return self._exchange_map

    def parse_values(self):
        model = self.community.comm_model
        reaction_map = self.community.reaction_map

        self.growth = self.values[self.community.biomass]

        self.abundance = {}
        for org_id, organism in self.community.organisms.items():
            growth_i = self.community.biomasses[org_id]
            self.abundance[org_id] = self.values[growth_i] / self.growth

        self.exchange = {r_id: self.values[r_id]
                         for r_id in model.get_exchange_reactions()}

        self.internal = {}
        self.normalized = {}

        for org_id, organism in self.community.organisms.items():

            abundance = self.abundance[org_id]

            fluxes = {r_id: self.values[reaction_map[(org_id, r_id)]]
                      for r_id in organism.reactions
                      if (org_id, r_id) in reaction_map}

            rates = {r_id: fluxes[r_id] / abundance if abundance > 0 else 0
                     for r_id in organism.reactions if r_id in fluxes}

            self.internal[org_id] = fluxes
            self.normalized[org_id] = rates

    # calculate overall exchanges (organism x metabolite) -> rate

    def compute_exchanges(self):
        model = self.community.comm_model
        reaction_map = self.community.reaction_map
        exchanges = {}

        for m_id in model.get_external_metabolites():

            for org_id, organism in self.community.organisms.items():
                rate = 0
                if m_id not in organism.metabolites:
                    continue

                for r_id in organism.get_metabolite_reactions(m_id):
                    if (org_id, r_id) not in reaction_map:
                        continue

                    flux = self.values[reaction_map[(org_id, r_id)]]

                    if flux != 0:
                        coeff = organism.reactions[r_id].stoichiometry[m_id]
                        rate += coeff*flux

                if rate != 0:
                    exchanges[(org_id, m_id)] = rate

        return exchanges

    def cross_feeding(self, as_df=False, abstol=1e-6):
        exchanges = self.compute_exchanges()
        cross_all = []

        for m_id in self.community.merged_model.get_external_metabolites():
            r_out = {x: r for (x, m), r in exchanges.items() if m == m_id and r > abstol}
            r_in = {x: -r for (x, m), r in exchanges.items() if m == m_id and -r > abstol}

            total_in = sum(r_in.values())
            total_out = sum(r_out.values())
            total = max(total_in, total_out)

            if total_in > total_out:
                r_out[None] = total_in - total_out
            if total_out > total_in:
                r_in[None] = total_out - total_in

            cross = [(o1, o2, m_id, r1 * r2 / total) for o1, r1 in r_out.items() for o2, r2 in r_in.items()]
            cross_all.extend(cross)

        if as_df:
            from pandas import DataFrame
            cross_all = DataFrame(cross_all, columns=["donor", "receiver", "compound", "rate"])

        return cross_all

    def mass_flow(self, element=None, as_df=False, abstol=1e-6):

        def get_mass(x):
            met = self.community.comm_model.get_metabolite(x)
            formula = x.formula
            mw = molecular_weight(formula, element=element)
            return 0.001 * mw

        entities = list(self.community.organisms) + [None]
        flow = {(o1, o2): 0 for o1 in entities for o2 in entities}

        for o1, o2, m_id, rate in self.cross_feeding():
            flow[(o1, o2)] += get_mass(m_id) * rate

        flow = {key: val for key, val in flow.items() if val > abstol}

        if as_df:
            from pandas import DataFrame
            flow = [(o1, o2, val) for (o1, o2), val in flow.items()]
            flow = DataFrame(flow, columns=["donor", "receiver", "flow"])

        return flow

    def print_external_fluxes(self, pattern=None, sort=False, abstol=1e-9):
        print_values(self.exchange, pattern=pattern, sort=sort, abstol=abstol)

    def print_internal_fluxes(self, org_id, normalized=False, pattern=None, sort=False, abstol=1e-9):
        if normalized:
            print_values(self.normalized[org_id], pattern=pattern, sort=sort, abstol=abstol)
        else:
            print_values(self.internal[org_id], pattern=pattern, sort=sort, abstol=abstol)

    def print_external_balance(self, m_id, sort=False, percentage=False, equations=False, abstol=1e-9):

        #print_balance(self.values, m_id, self.community.merged_model, sort=sort, percentage=percentage, equations=equations,
        #              abstol=abstol)
        return None

    def print_exchanges(self, m_id=None, abstol=1e-9):

        model = self.community.comm_model
        exchange_rxns = set(model.get_exchange_reactions())

        if m_id:
            mets = [m_id]
        else:
            mets = model.get_external_metabolites()

        for m_id in mets:

            entries = []

            ext_rxns = set(model.get_metabolite_reactions(m_id)) & exchange_rxns

            for r_id in ext_rxns:

                flux = self.values[r_id]
                coeff = model.reactions[r_id].stoichiometry[m_id]
                rate = coeff*flux

                if rate > abstol:
                    entries.append(('=> *   ', "in", rate))
                elif rate < -abstol:
                    entries.append(('   * =>', "out", rate))

            for org_id in self.community.organisms:

                if (org_id, m_id) not in self.exchange_map:
                    continue

                rate = self.exchange_map[(org_id, m_id)]
                if rate > abstol:
                    entries.append(('O --> *', org_id, rate))
                elif rate < -abstol:
                    entries.append(('* --> O', org_id, rate))

            if entries:
                print(m_id)
                entries.sort(key=lambda x: x[2])

                for (sense, org_id, rate) in entries:
                    print(f'[ {sense} ] {org_id:<12} {rate:< 10.6g}')



import copy
import warnings
from abc import ABC, abstractmethod
from enum import Enum
import numpy as np
from ..optimization.ea import Solution, filter_duplicates
from ..simulation import get_simulator
from ..util.constants import EAConstants, ModelConstants
from ..util.process import get_fevaluator


class Strategy(Enum):
    KO = 'KO'
    OU = 'OU'

    def __eq__(self, other):
        """Overrides equal to enable string name comparison.
        Allows to seamlessly use:
            Strategy.KO = Strategy.KO
            Strategy.KO = 'KO'.
        """
        if isinstance(other, Strategy):
            return super().__eq__(other)
        elif isinstance(other, str):
            return self.name == other
        else:
            return False

    def __hash__(self):
        return hash(self.name)


class KOBounder(object):
    """
    A bounder of possible indexes in an enumeration
    """

    def __init__(self, lower_bound, upper_bound):

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.range = self.upper_bound - self.lower_bound + 1

    # make it callable
    def __call__(self, candidate, args):
        bounded_candidate = set()
        for val in candidate:
            if val > self.upper_bound or val < self.lower_bound:
                val = val % self.range + self.lower_bound
            bounded_candidate.add(val)
        return bounded_candidate


class OUBounder(object):
    """
    A bounder for (int,int) representations
    """

    def __init__(self, lower_bound, upper_bound):
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.range = [self.upper_bound[i] - self.lower_bound[i] +
                      1 for i in range(len(self.lower_bound))]

    def __call__(self, candidate, args):
        bounded_candidate = set()
        for idx, lv in candidate:
            if idx > self.upper_bound[0] or idx < self.lower_bound[0]:
                idx = idx % self.range[0] + self.lower_bound[0]
            if lv > self.upper_bound[1] or idx < self.lower_bound[1]:
                lv = lv % self.range[1] + self.lower_bound[1]
            bounded_candidate.add((idx, lv))
        return bounded_candidate


class AbstractProblem(ABC):
    """
    Base class for optimization problems.

    :param model: A Metabolic model.
    :param list fevaluation: A list of callable EvaluationFunctions.
    :param kwargs: Additional parameters dictionary
    """

    def __init__(self, model, fevaluation=None, **kwargs):
        self.model = model
        self.fevaluation = [] if fevaluation is None else fevaluation
        self.number_of_objectives = len(self.fevaluation)

        # simulation context : defines the simulations environment
        self._reset_solver = kwargs.get('reset_solver', ModelConstants.RESET_SOLVER)
        self._simul = None
        # The target product reaction id may be specified when optimizing for a single product.
        # Only required for probabilistic modification targeting.
        self.product = kwargs.get('product', None)
        # Environmental conditions
        self.environmental_conditions = kwargs.get('envcond', None)
        # Additional persistent constraints
        self.persistent_constraints = kwargs.get('constraints', None)
        # Reference reaction fluxes
        self._reference = None
        # solution size
        self.candidate_min_size = kwargs.get(
            'candidate_min_size', EAConstants.MIN_SOLUTION_SIZE)
        self.candidate_max_size = kwargs.get(
            'candidate_max_size', EAConstants.MAX_SOLUTION_SIZE)
        # non target
        self.non_target = kwargs.get('non_target', None)
        # targets
        # If not provided, targets are build in the context of the problem.
        # Objectives are not automatically removed from the targets... this should be a user concern!?
        self._trg_list = kwargs.get('target', None)
        # the EA representation bounds
        self._bounder = None
        # scaling factor
        self.scalefactor = kwargs.get('scalefactor', None)
        # required simulations
        methods = []
        for f in self.fevaluation:
            methods.extend(f.required_simulations())
        self.methods = list(set(methods))
        # problem specific EA operators
        self.operators = None
        self.operators_param = None

    @abstractmethod
    def generator(self, random, **kwargs):
        """The generator function for the problem."""
        raise NotImplementedError

    @abstractmethod
    def encode(self, candidate):
        """The generator function for the problem."""
        raise NotImplementedError

    @abstractmethod
    def decode(self, candidate):
        """The decoder function for the problem."""
        raise NotImplementedError

    @abstractmethod
    def solution_to_constraints(self, solution):
        """Converts a decoded solution to metabolict constraints."""
        raise NotImplementedError

    def get_name(self):
        """The problem name."""
        return self.__class__.__name__

    def pre_process(self):
        """ Defines pre processing tasks
        """
        self.target_list
        self.reset_simulator()

    @property
    def simulator(self):
        if self._simul is None:
            self._simul = get_simulator(
                self.model, envcond=self.environmental_conditions,
                constraints=self.persistent_constraints,
                reference=self._reference, reset_solver=self._reset_solver)
        return self._simul

    def simulate(self, *args, **kwargs):
        return self._simul.simulate(*args, **kwargs)

    def reset_simulator(self):
        self._simul = None

    def __str__(self):
        if self.number_of_objectives > 1:
            return '{0} ({1}  objectives)'.format(self.__class__.__name__, self.number_of_objectives)
        else:
            return '{0} '.format(self.__class__.__name__)

    def __repr__(self):
        return self.__class__.__name__

    @property
    def target_list(self):
        "list of modification targets"
        if self._trg_list is None:
            self._build_target_list()
            if not self._trg_list:
                raise RuntimeError("Could not build target list.")
        return self._trg_list

    @abstractmethod
    def _build_target_list(self):
        raise NotImplementedError

    def get_constraints(self, solution):
        """
        :returns: The constrainst enconded into an individual.
        """
        return self.solution_to_constraints(self.decode(solution.candidate))

    def get_environmental_conditions(self):
        return self.environmental_conditions

    def get_persistent_constraints(self):
        return self.persistent_constraints

    def evaluate_solution(self, solution, decode=True):
        """
        Evaluates a single solution, a list of constraints.

        :param solution: The solution to be evaluated.
        :param decode: If the solution needs to be decoded.
        :returns: A list of fitness.
        """
        decoded = {}
        # decoded constraints
        if decode:
            decoded = self.decode(solution)
            constraints = self.solution_to_constraints(decoded)
        else:
            constraints = solution

        # pre simulation
        simulation_results = dict()
        try:
            p = []
            for method in self.methods:
                simulation_result = self.simulator.simulate(
                    constraints=constraints, method=method, scalefactor=self.scalefactor)
                simulation_results[method] = simulation_result
            # apply the evaluation function(s)
            for f in self.fevaluation:
                v = f(simulation_results, decoded, scalefactor=self.scalefactor)
                p.append(v)
        except Exception as e:
            p = []
            for f in self.fevaluation:
                p.append(f.worst_fitness)
            if EAConstants.DEBUG:
                warnings.warn(f"Solution couldn't be evaluated [{e}]\n {constraints}")
        return p

    @property
    def is_maximization(self):
        return all([f.maximize for f in self.fevaluation])

    def simplify(self, solution, tolerance=1e-6, ):
        """
        Simplify a solution by removing the modification that do not affect the final fitness value.
        Two solutions are considered different if the maximum allowed difference between objective values is exceeded.
        Tolerance may be defined by a single float value, or per objective by setting a list of floats of size equal
        to the number of objectives.

        :param solution: the solution to be simplified
        :param float tolerance: The maximum allowed difference between objective values.
        :returns: A list of simplified solutions.

        """

        enc_values = self.encode(solution.values)
        fitness = self.evaluate_solution(enc_values)
        one_to_remove = {}
        # single removal
        for entry in enc_values:
            simul_enc_values = copy.copy(enc_values)
            simul_enc_values.remove(entry)
            fit = self.evaluate_solution(simul_enc_values)
            diff = np.abs(np.array(fit) - np.array(fitness))
            is_equal = False
            if isinstance(tolerance, float):
                is_equal = np.all(diff <= tolerance)
            else:
                is_equal = np.all(diff <= np.array(tolerance))
            if is_equal:
                one_to_remove[entry] = fit

        simul_enc_values = copy.copy(enc_values)
        for entry in one_to_remove.keys():
            simul_enc_values.remove(entry)

        # test all simultaneous removal
        fit = self.evaluate_solution(simul_enc_values)
        diff = np.abs(np.array(fit) - np.array(fitness))
        is_equal = False
        if isinstance(tolerance, float):
            is_equal = np.all(diff <= tolerance)
        else:
            is_equal = np.all(diff <= np.array(tolerance))

        if is_equal:
            v = self.decode(simul_enc_values)
            c = self.solution_to_constraints(v)
            simplification = Solution(v, fitness, c)
            return [simplification]
        else:
            res = []
            for entry, fit in one_to_remove.items():
                simul_enc_values = copy.copy(enc_values)
                simul_enc_values.remove(entry)
                v = self.decode(simul_enc_values)
                c = self.solution_to_constraints(v)
                simplification = Solution(v, fitness, c)
                res.append(simplification)
            return res

    def simplify_population(self, population, n_cpu=1):
        """Simplifies a population of solutions

        Args:
            population (list): List of mewpy.optimization.ea.Solution
            n_cpu (int): Number of CPUs.
        Returns:
            list: Simplified population
        """
        pop = []
        for solution in population:
            try:
                res = self.simplify(solution)
                pop.extend(res)
            except Exception:
                pop.append(solution)
        return filter_duplicates(pop)


class AbstractKOProblem(AbstractProblem):

    def __init__(self, model, fevaluation=None, **kwargs):
        """
        Base class for Knockout optimization problems.

        :param model: The constraint metabolic model.
        :param list fevaluation: A list of callable EvaluationFunctions.

        """
        super(AbstractKOProblem, self).__init__(
            model, fevaluation=fevaluation, **kwargs)
        self.strategy = Strategy.KO

    def decode(self, candidate):
        decoded = dict()
        for idx in candidate:
            try:
                decoded[self.target_list[idx]] = 0
            except IndexError:
                raise IndexError("Index out of range: {} from {}".format(
                    idx, len(self.target_list[idx])))
        return decoded

    def encode(self, candidate):
        """
        Translates a candidate solution in problem specific representation to
        an iterable of ids, or (ids, folds).

        :param candidate: The candidate representation.
        """
        return set([self.target_list.index(k) for k in candidate])

    def solution_to_constraints(self, decoded_candidate):
        """
        Converts a candidate, a dictionary of reactions, into a dictionary of constraints.
        This is problem specific. By default return the decoded candidate.
        """
        return decoded_candidate

    @property
    def bounder(self):
        """
        The KO list index bounder
        """
        if self._bounder is None:
            max = len(self.target_list) - 1
            self._bounder = KOBounder(0, max)
        return self._bounder

    def generator(self, random, **kwargs):
        """
        Generates a solution, a random int set with length in range min_solution_size to max_solution_size.
        """

        if self.candidate_min_size == self.candidate_max_size:
            solution_size = self.candidate_min_size
        else:
            solution_size = random.randint(
                self.candidate_min_size, self.candidate_max_size)

        solution = set(random.sample(range(len(self.target_list)), solution_size))
        return solution


class AbstractOUProblem(AbstractProblem):
    """ Base class for Over/Under expression optimization problems
    """

    def __init__(self, model, fevaluation=None, **kwargs):
        """
        :param model: The constraint metabolic model.
        :param list fevaluation: A list of callable EvaluationFunctions.

        Optional:

        :param dic reference: Dictionary of flux values to be used in the over/under expression values computation.
        :param list levels: Over/under expression levels (Default EAConstants.LEVELS).
        :param boolean twostep: If deletions should be applied before identifiying reference flux values.
        :param dict partial_solution: A partial solution to be appended to any other solution
        """
        super(AbstractOUProblem, self).__init__(
            model, fevaluation=fevaluation, **kwargs)
        self.strategy = Strategy.OU
        self.levels = kwargs.get('levels', EAConstants.LEVELS)
        self._reference = kwargs.get('reference', None)
        self.twostep = kwargs.get('twostep', False)
        self._partial_solution = kwargs.get('partial_solution', dict())

    def decode(self, candidate):
        """The decoder function for the problem. Needs to be implemented by extending classes."""
        decoded = self._partial_solution.copy()
        for idx, lv_idx in candidate:
            try:
                rxn = self.target_list[idx]
                lv = self.levels[lv_idx]
                decoded[rxn] = lv
            except IndexError:
                raise IndexError("Index out of range")
        return decoded

    def encode(self, candidate):
        """
        Translates a candidate solution in problem specific representation to
        an iterable of ids, or (ids, folds).

        :param iterable candidate: The candidate representation.
        :returns: a list of index tupple (modification_target_index,level_index). The indexes are
                  problem dependent.
        """

        return set([(self.target_list.index(k), self.levels.index(lv))
                    for k, lv in candidate.items()])

    def solution_to_constraints(self, decoded_candidate):
        """
        Decodes a candidate, a dictionary of reactions, into a dictionary of constraints.
        This is problem specific. By default return the decoded candidate.
        """
        return decoded_candidate

    @property
    def bounder(self):
        """
        The list and levels index bounder

        :returns: a OUBounder object.
        """
        if self._bounder is None:
            max_idx = len(self.target_list) - 1
            max_lv = len(self.levels) - 1
            self._bounder = OUBounder([0, 0], [max_idx, max_lv])
        return self._bounder

    def generator(self, random, **kwargs):
        """
        Generates a solution, a random (int,int) set with length in range min_solution_size to max_solution_size.

        :param random: A random number generator. Needs to implement uniform and randint generators methods.
        :param dict args: A dictionary of additional parameters.
        :returns: A new solution.
        """

        if self.candidate_min_size == self.candidate_max_size:
            solution_size = self.candidate_min_size
        else:
            solution_size = random.randint(
                self.candidate_min_size, self.candidate_max_size)

        keys = random.sample(range(len(self.target_list)), solution_size)
        values = random.choices(range(len(self.levels)), k=solution_size)
        solution = set(zip(keys, values))
        return solution

    @property
    def reference(self):
        """
        :returns: A dictionary of reference flux values.
        """
        if not self._reference:
            self._reference = self.simulator.reference
        return self._reference

    def ou_constraint(self, level, wt):
        """ Computes the bounds for a reaction.

        :param float level: The expression level for the reaction.
        :param float wt: The reference reaction flux.
        :returns: A tupple, flux bounds for the reaction.
        """
        if level > 1:
            if wt > 0:
                return (level * wt, ModelConstants.REACTION_UPPER_BOUND)
            elif wt < 0:
                return (-1 * ModelConstants.REACTION_UPPER_BOUND, level * wt)
        else:
            if wt > 0:
                return (0, level * wt)
            elif wt < 0:
                return (level * wt, 0)

    def reaction_constraints(self, rxn, lv, reference):
        """
        Converts a (reaction, level) pair into a constraint
        If a reaction is reversible, the direction with no or less wild type flux
        is knocked out.

        :param rxn: The reaction identifier.
        :param lv: the wild type multiplier factor. The reaction bounds are altered accordingly.
        :returns: A dictionary of reaction constraints.
        """
        constraints = {}
        fluxe_wt = reference[rxn]
        rev_rxn = self.simulator.reverse_reaction(rxn)
        if lv == 0:
            # KO constraint
            constraints[rxn] = (0, 0)
        elif lv == 1 or fluxe_wt == 0:
            # No contraint is applyed
            pass
        elif rev_rxn is None or rev_rxn == rxn:
            # if there is no reverse reaction
            constraints[rxn] = self.ou_constraint(lv, fluxe_wt)
        else:
            # there's a reverse reaction...
            # one of the two reactions needs to be KO, the one with no flux in the wt.
            rev_fluxe_wt = reference[rev_rxn]
            if abs(fluxe_wt) >= abs(rev_fluxe_wt):
                ko_rxn, ou_rxn, fwt = rev_rxn, rxn, fluxe_wt
            else:
                rxn, rev_rxn, rev_fluxe_wt
            constraints[ko_rxn] = (0, 0)
            constraints[ou_rxn] = self.ou_constraint(lv, fwt)
        return constraints



from enum import Enum

from .fba import FBA
from .pfba import pFBA
from .rfba import RFBA
from .srfba import SRFBA
from .prom import PROM, target_regulator_interaction_probability
from .coregflux import CoRegFlux, predict_gene_expression
from .metabolic_analysis import slim_fba, slim_pfba, fva, single_gene_deletion, single_reaction_deletion
from .regulatory_analysis import regulatory_truth_table
from .integrated_analysis import (slim_rfba, slim_srfba, slim_prom, slim_coregflux,
                                  ifva, isingle_regulator_deletion, isingle_reaction_deletion, isingle_gene_deletion,
                                  find_conflicts)


class Analysis(Enum):
    """
    Enumeration of the available analysis methods.
    """
    FBA = FBA
    pFBA = pFBA
    RFBA = RFBA
    SRFBA = SRFBA
    PROM = PROM
    CoRegFlux = CoRegFlux

    @classmethod
    def has_analysis(cls, analysis: str) -> bool:
        """
        Check if the analysis is available.

        :param analysis: The name of the analysis
        :return: True if the analysis is available, False otherwise
        """
        return analysis in cls.__members__

    @classmethod
    def get(cls, analysis: str, default=FBA) -> 'Analysis':
        """
        Get the analysis class.

        :param analysis: the analysis name
        :param default: the default analysis to return if the analysis is not available
        :return: the analysis class
        """
        try:
            return cls[analysis]

        except KeyError:

            return default
